#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from earthkit.data import from_source, settings
from earthkit.data.core.temporary import temp_directory
from earthkit.data.testing import earthkit_file, network_off


@pytest.mark.skipif(  # TODO: fix
    sys.platform == "win32",
    reason="file:// not working on Windows yet",
)
def test_url_file_source():
    filename = os.path.abspath(earthkit_file("docs/examples/test.nc"))
    s = from_source("url", f"file://{filename}")
    assert len(s) == 2


def test_url_source_1():
    from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test.grib",
    )


def test_url_source_check_out_of_date():
    def load():
        from_source(
            "url",
            "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test.grib",
        )

    with temp_directory() as tmpdir:
        with settings.temporary():
            settings.set("user-cache-directory", tmpdir)
            load()

            settings.set("check-out-of-date-urls", False)
            with network_off():
                load()


def test_url_source_2():
    from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
    )


def test_url_source_3():
    from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test.nc",
    )


def test_url_source_tar():
    ds = from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test_gribs.tar",
    )
    assert len(ds) == 6


def test_part_url():
    ds = from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
    )

    ds = from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
        parts=((0, 4),),
    )

    assert os.path.getsize(ds.path) == 4

    with open(ds.path, "rb") as f:
        assert f.read() == b"BUFR"

    ds = from_source(
        "url",
        "https://get.ecmwf.int/repository/test-data/earthkit-data/test-data/temp.bufr",
        parts=((0, 10), (50, 10), (60, 10)),
    )

    print(ds.path)

    assert os.path.getsize(ds.path) == 30

    with open(ds.path, "rb") as f:
        assert f.read()[:4] == b"BUFR"


@pytest.mark.skipif(  # TODO: fix
    sys.platform == "win32",
    reason="file:// not working on Windows yet",
)
def test_url_part_file_source():
    filename = os.path.abspath(earthkit_file("docs/examples/test.grib"))
    ds = from_source(
        "url",
        f"file://{filename}",
        parts=[
            (0, 4),
            (522, 4),
            (526, 4),
            (1048, 4),
        ],
    )

    assert os.path.getsize(ds.path) == 16

    with open(ds.path, "rb") as f:
        assert f.read() == b"GRIB7777GRIB7777"


if __name__ == "__main__":
    test_part_url()
    # from earthkit.data.testing import main

    # main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file, earthkit_test_data_file


def check_array(v, shape=None, first=None, last=None, meanv=None, eps=1e-3):
    assert v.shape == shape
    assert np.isclose(v[0], first, eps)
    assert np.isclose(v[-1], last, eps)
    assert np.isclose(v.mean(), meanv, eps)


def repeat_list_items(items, count):
    return sum([[x] * count for x in items], [])


@pytest.mark.parametrize(
    "key,expected_value",
    [
        ("shortName", "2t"),
        ("shortName:s", "2t"),
        ("shortName:str", "2t"),
        ("centre", "ecmf"),
        ("centre:l", 98),
        ("level", 0),
        ("level:l", 0),
        ("level:int", 0),
        (["shortName"], ["2t"]),
        (["shortName", "level"], ["2t", 0]),
        (("shortName"), "2t"),
        (("shortName", "level"), ("2t", 0)),
    ],
)
def test_grib_metadata_grib(key, expected_value):
    f = from_source("file", earthkit_test_data_file("test_single.grib"))
    sn = f.metadata(key)
    assert sn == [expected_value]
    sn = f[0].metadata(key)
    assert sn == expected_value


@pytest.mark.parametrize(
    "key,astype,expected_value",
    [
        ("shortName", str, "2t"),
        ("shortName", None, "2t"),
        ("centre", None, "ecmf"),
        ("centre", str, "ecmf"),
        ("centre", int, 98),
        ("level", None, 0),
        ("level", str, "0"),
        ("level", int, 0),
    ],
)
def test_grib_metadata_astype_1(key, astype, expected_value):
    f = from_source("file", earthkit_test_data_file("test_single.grib"))
    sn = f.metadata(key, astype=astype)
    assert sn == [expected_value]
    sn = f[0].metadata(key, astype=astype)
    assert sn == expected_value


@pytest.mark.parametrize(
    "key,expected_value",
    [
        ("shortName", ["t", "u", "v"] * 6),
        ("shortName:s", ["t", "u", "v"] * 6),
        ("shortName:str", ["t", "u", "v"] * 6),
        ("level", repeat_list_items([1000, 850, 700, 500, 400, 300], 3)),
        ("level:l", repeat_list_items([1000, 850, 700, 500, 400, 300], 3)),
        ("level:int", repeat_list_items([1000, 850, 700, 500, 400, 300], 3)),
    ],
)
def test_grib_metadata_18(key, expected_value):
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))
    sn = f.metadata(key)
    assert sn == expected_value


@pytest.mark.parametrize(
    "key,astype,expected_value",
    [
        ("shortName", str, ["t", "u", "v"] * 6),
        ("shortName", None, ["t", "u", "v"] * 6),
        (
            "level",
            int,
            repeat_list_items([1000, 850, 700, 500, 400, 300], 3),
        ),
        (
            "level",
            None,
            repeat_list_items([1000, 850, 700, 500, 400, 300], 3),
        ),
    ],
)
def test_grib_metadata_astype_18(key, astype, expected_value):
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))
    sn = f.metadata(key, astype=astype)
    assert sn == expected_value


@pytest.mark.parametrize(
    "key,expected_value",
    [
        ("max", 307.18560791015625),
        ("max:d", 307.18560791015625),
        ("max:float", 307.18560791015625),
    ],
)
def test_grib_metadata_double_1(key, expected_value):
    f = from_source("file", earthkit_test_data_file("test_single.grib"))
    r = f.metadata(key)
    assert len(r) == 1
    assert np.isclose(r[0], expected_value)


@pytest.mark.parametrize(
    "key",
    [
        ("max"),
        ("max:d"),
        ("max:float"),
    ],
)
def test_grib_metadata_double_18(key):
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))

    ref = [
        320.5641784667969,
        17.713119506835938,
        11.833480834960938,
        304.53916931152344,
        27.10162353515625,
        12.660964965820312,
        287.26531982421875,
        28.145523071289062,
        15.6385498046875,
        271.8430633544922,
        36.74000549316406,
        15.009902954101562,
        264.00323486328125,
        46.213775634765625,
        23.949615478515625,
        250.6531524658203,
        58.45549011230469,
        36.92034912109375,
    ]

    r = f.metadata(key)
    np.testing.assert_allclose(r, ref, 0.001)


@pytest.mark.parametrize(
    "key,astype",
    [
        ("max", None),
        ("max", float),
    ],
)
def test_grib_metadata_double_astype_18(key, astype):
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))

    ref = [
        320.5641784667969,
        17.713119506835938,
        11.833480834960938,
        304.53916931152344,
        27.10162353515625,
        12.660964965820312,
        287.26531982421875,
        28.145523071289062,
        15.6385498046875,
        271.8430633544922,
        36.74000549316406,
        15.009902954101562,
        264.00323486328125,
        46.213775634765625,
        23.949615478515625,
        250.6531524658203,
        58.45549011230469,
        36.92034912109375,
    ]

    r = f.metadata(key, astype=astype)
    np.testing.assert_allclose(r, ref, 0.001)


def test_grib_get_long_array_1():
    f = from_source(
        "file", earthkit_test_data_file("rgg_small_subarea_cellarea_ref.grib")
    )
    assert len(f) == 1
    pl = f.metadata("pl")
    assert len(pl) == 1
    pl = pl[0]
    assert isinstance(pl, np.ndarray)
    assert len(pl) == 73
    assert pl[0] == 24
    assert pl[1] == 28
    assert pl[20] == 104
    assert pl[72] == 312


def test_grib_get_double_array_values_1():
    f = from_source("file", earthkit_test_data_file("test_single.grib"))
    v = f.metadata("values")
    assert len(v) == 1
    v = v[0]
    assert isinstance(v, np.ndarray)
    eps = 1e-5
    check_array(
        v,
        (84,),
        first=260.43560791015625,
        last=227.18560791015625,
        meanv=274.36566743396577,
        eps=eps,
    )


def test_grib_get_double_array_values_18():
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))
    v = f.metadata("values")
    assert isinstance(v, list)
    assert len(v) == 18
    for d in v:
        assert isinstance(d, np.ndarray)
        assert d.shape == (84,)

    eps = 1e-5
    check_array(
        v[0],
        (84,),
        first=272.5642,
        last=240.56417846679688,
        meanv=279.70703560965404,
        eps=eps,
    )

    check_array(
        v[15],
        (84,),
        first=226.6531524658203,
        last=206.6531524658203,
        meanv=227.84362865629652,
        eps=eps,
    )


def test_grib_get_double_array_1():
    f = from_source("file", earthkit_test_data_file("ml_data.grib"))[0]
    # f is now a field!
    v = f.metadata("pv")
    assert isinstance(v, np.ndarray)
    assert len(v) == 276
    assert np.isclose(v[0], 0.0)
    assert np.isclose(v[1], 2.0003650188446045)
    assert np.isclose(v[20], 316.4207458496094)
    assert np.isclose(v[275], 1.0)


def test_grib_get_double_array_18():
    f = from_source("file", earthkit_test_data_file("ml_data.grib"))
    v = f.metadata("pv")
    assert isinstance(v, list)
    assert len(v) == 36
    for row in v:
        assert isinstance(row, np.ndarray)
        assert len(row) == 276

    eps = 0.001
    assert np.isclose(v[0][1], 2.0003650188446045, eps)
    assert np.isclose(v[0][20], 316.4207458496094, eps)
    assert np.isclose(v[17][1], 2.0003650188446045, eps)
    assert np.isclose(v[17][20], 316.4207458496094, eps)


def test_grib_metadata_type_qualifier():
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))[0:4]

    # to str
    r = f.metadata("centre:s")
    assert r == ["ecmf", "ecmf", "ecmf", "ecmf"]
    r = f.metadata("centre:str")
    assert r == ["ecmf", "ecmf", "ecmf", "ecmf"]
    r = f.metadata("level:s")
    assert r == ["1000", "1000", "1000", "850"]
    r = f.metadata("level:str")
    assert r == ["1000", "1000", "1000", "850"]

    # to int
    r = f.metadata("centre:l")
    assert r == [98, 98, 98, 98]
    r = f.metadata("centre:int")
    assert r == [98, 98, 98, 98]
    r = f.metadata("level:d")
    assert r == [1000, 1000, 1000, 850]
    r = f.metadata("level:int")
    assert r == [1000, 1000, 1000, 850]

    # to float
    r = f.metadata("centre:d")
    assert np.allclose(np.array(r), np.array([98.0, 98.0, 98.0, 98.0]))
    assert all(isinstance(x, float) for x in r)
    r = f.metadata("centre:float")
    assert np.allclose(np.array(r), np.array([98.0, 98.0, 98.0, 98.0]))
    assert all(isinstance(x, float) for x in r)
    r = f.metadata("level:d")
    assert np.allclose(np.array(r), np.array([1000.0, 1000.0, 1000.0, 850.0]))
    assert all(isinstance(x, float) for x in r)
    r = f.metadata("level:float")
    assert np.allclose(np.array(r), np.array([1000.0, 1000.0, 1000.0, 850.0]))
    assert all(isinstance(x, float) for x in r)


def test_grib_metadata_astype():
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))[0:4]

    # to str
    r = f.metadata("centre", astype=None)
    assert r == ["ecmf", "ecmf", "ecmf", "ecmf"]
    r = f.metadata("centre", astype=str)
    assert r == ["ecmf", "ecmf", "ecmf", "ecmf"]
    r = f.metadata("level", astype=str)
    assert r == ["1000", "1000", "1000", "850"]

    # to int
    r = f.metadata("centre", astype=int)
    assert r == [98, 98, 98, 98]
    r = f.metadata("level", astype=int)
    assert r == [1000, 1000, 1000, 850]

    # to float
    r = f.metadata("level", astype=float)
    assert np.allclose(np.array(r), np.array([1000.0, 1000.0, 1000.0, 850.0]))
    assert all(isinstance(x, float) for x in r)

    # multi
    r = f.metadata(["level", "cfVarName"], astype=(int, None))
    assert r == [[1000, "t"], [1000, "u"], [1000, "v"], [850, "t"]]
    r = f.metadata(["level", "cfVarName"], astype=str)
    assert r == [["1000", "t"], ["1000", "u"], ["1000", "v"], ["850", "t"]]

    # non matching astype
    with pytest.raises(ValueError):
        f.metadata(["level", "cfVarName", "centre"], astype=(int, None))


def test_grib_metadata_generic():
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))[0:4]

    sn = f.metadata("shortName")
    assert sn == ["t", "u", "v", "t"]
    sn = f.metadata(["shortName"])
    assert sn == [["t"], ["u"], ["v"], ["t"]]
    lg = f.metadata("level", "cfVarName")
    assert lg == [(1000, "t"), (1000, "u"), (1000, "v"), (850, "t")]
    lg = f.metadata(["level", "cfVarName"])
    assert lg == [[1000, "t"], [1000, "u"], [1000, "v"], [850, "t"]]
    lg = f.metadata("level", "cfVarName")
    assert lg == [(1000, "t"), (1000, "u"), (1000, "v"), (850, "t")]

    # single fieldlist
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))
    f = f.sel(count=[1])
    lg = f.metadata(["level", "cfVarName"])
    assert lg == [[1000, "t"]]

    # single field
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))[0]
    lg = f.metadata(["level", "cfVarName"])
    assert lg == [1000, "t"]


def test_grib_metadata_missing_value():
    f = from_source("file", earthkit_test_data_file("ml_data.grib"))

    with pytest.raises(KeyError):
        f[0].metadata("scaleFactorOfSecondFixedSurface")

    v = f[0].metadata("scaleFactorOfSecondFixedSurface", default=None)
    assert v is None


def test_grib_metadata_missing_key():
    f = from_source("file", earthkit_examples_file("test.grib"))

    with pytest.raises(KeyError):
        f[0].metadata("_badkey_")

    v = f[0].metadata("__badkey__", default=0)
    assert v == 0


def test_grib_metadata_namespace():
    f = from_source("file", earthkit_examples_file("test6.grib"))

    r = f[0].metadata(namespace="vertical")
    ref = {"level": 1000, "typeOfLevel": "isobaricInhPa"}
    assert r == ref

    r = f[0].metadata(namespace=["vertical", "time"])
    ref = {
        "vertical": {"typeOfLevel": "isobaricInhPa", "level": 1000},
        "time": {
            "dataDate": 20180801,
            "dataTime": 1200,
            "stepUnits": 1,
            "stepType": "instant",
            "stepRange": "0",
            "startStep": 0,
            "endStep": 0,
            "validityDate": 20180801,
            "validityTime": 1200,
        },
    }
    assert r == ref

    r = f[0].metadata(namespace=None)
    assert isinstance(r, dict)
    assert len(r) == 186
    assert r["level"] == 1000
    assert r["stepType"] == "instant"

    r = f[0].metadata(namespace=[None])
    assert isinstance(r, dict)
    assert len(r) == 186
    assert r["level"] == 1000
    assert r["stepType"] == "instant"

    r = f[0].metadata(namespace="")
    assert isinstance(r, dict)
    assert len(r) == 186
    assert r["level"] == 1000
    assert r["stepType"] == "instant"

    r = f[0].metadata(namespace=[""])
    assert isinstance(r, dict)
    assert len(r) == 186
    assert r["level"] == 1000
    assert r["stepType"] == "instant"

    ref = {
        "geography",
        "statistics",
        "vertical",
        "time",
        "parameter",
        "mars",
        "ls",
        "default",
    }
    r = f[0].metadata(namespace=all)
    assert isinstance(r, dict)
    assert set(r.keys()) == ref

    r = f[0].metadata(namespace=[all])
    assert isinstance(r, dict)
    assert set(r.keys()) == ref

    with pytest.raises(ValueError) as excinfo:
        r = f[0].metadata("level", namespace=["vertical", "time"])
    assert "must be a str when key specified" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        r = f[0].metadata("level", namespace=["vertical", "time"])
    assert "must be a str when key specified" in str(excinfo.value)


def test_grib_datetime():
    s = from_source("file", earthkit_examples_file("test.grib"))

    ref = {
        "base_time": [datetime.datetime(2020, 5, 13, 12)],
        "valid_time": [datetime.datetime(2020, 5, 13, 12)],
    }
    assert s.datetime() == ref

    s = from_source(
        "dummy-source",
        kind="grib",
        paramId=[129, 130],
        date=[19900101, 19900102],
        level=[1000, 500],
    )
    ref = {
        "base_time": [
            datetime.datetime(1990, 1, 1, 12, 0),
            datetime.datetime(1990, 1, 2, 12, 0),
        ],
        "valid_time": [
            datetime.datetime(1990, 1, 1, 12, 0),
            datetime.datetime(1990, 1, 2, 12, 0),
        ],
    }
    assert s.datetime() == ref


def test_grib_valid_datetime():
    ds = from_source("file", earthkit_test_data_file("t_time_series.grib"))
    f = ds[4]

    assert f.metadata("valid_datetime") == datetime.datetime(2020, 12, 21, 18)


def test_message():
    f = from_source("file", earthkit_examples_file("test.grib"))
    v = f[0].message()
    assert len(v) == 526
    assert v[:4] == b"GRIB"
    v = f[1].message()
    assert len(v) == 526
    assert v[:4] == b"GRIB"


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging
import os
import threading
import time

import eccodes
import numpy as np

from earthkit.data.core.caching import CACHE, auxiliary_cache_file

LOG = logging.getLogger(__name__)

# For some reason, cffi can get stuck in the GC if that function
# needs to be called defined for the first time in a GC thread.
try:
    _h = eccodes.codes_new_from_samples(
        "regular_ll_pl_grib1", eccodes.CODES_PRODUCT_GRIB
    )
    eccodes.codes_release(_h)
except Exception:
    pass


class CodesMessagePositionIndex:
    VERSION = 1

    def __init__(self, path):
        self.path = path
        self.offsets = None
        self.lengths = None
        self._cache_file = None
        self._load()

    def __len__(self):
        return len(self.offsets)

    def _get_message_positions(self, path):
        raise NotImplementedError

    def _build(self):
        offsets = []
        lengths = []

        for offset, length in self._get_message_positions(self.path):
            offsets.append(offset)
            lengths.append(length)

        self.offsets = offsets
        self.lengths = lengths

    def _load(self):
        if CACHE.policy.use_message_position_index_cache():
            self._cache_file = auxiliary_cache_file(
                "message-index",
                self.path,
                content="null",
                extension=".json",
            )
            if not self._load_cache():
                self._build()
                self._save_cache()
        else:
            self._build()

    def _save_cache(self):
        if CACHE.policy.use_message_position_index_cache():
            try:
                with open(self._cache_file, "w") as f:
                    json.dump(
                        dict(
                            version=self.VERSION,
                            offsets=self.offsets,
                            lengths=self.lengths,
                        ),
                        f,
                    )
            except Exception:
                LOG.exception("Write to cache failed %s", self._cache_file)

    def _load_cache(self):
        if CACHE.policy.use_message_position_index_cache():
            try:
                with open(self._cache_file) as f:
                    c = json.load(f)
                    if not isinstance(c, dict):
                        return False

                    assert c["version"] == self.VERSION
                    self.offsets = c["offsets"]
                    self.lengths = c["lengths"]
                    return True
            except Exception:
                LOG.exception("Load from cache failed %s", self._cache_file)

        return False


class CodesHandle(eccodes.Message):
    MISSING_VALUE = np.finfo(np.float32).max
    KEY_TYPES = {"s": str, "l": int, "d": float}
    PRODUCT_ID = None

    def __init__(self, handle, path, offset):
        super().__init__(handle)
        self.path = path
        self.offset = offset

    @classmethod
    def from_sample(cls, name):
        return cls(eccodes.codes_new_from_samples(name, cls.PRODUCT_ID), None, None)

    @classmethod
    def _from_raw_handle(cls, handle):
        return cls(handle, None, None)

    # TODO: just a wrapper around the base class implementation to handle the
    # s,l,d qualifiers. Once these are implemented in the base class this method can
    # be removed. md5GridSection is also handled!
    def get(self, name, ktype=None, **kwargs):
        if ktype is None:
            key_name, _, key_type_str = name.partition(":")
            if key_type_str in CodesHandle.KEY_TYPES:
                name = key_name
                ktype = CodesHandle.KEY_TYPES[key_type_str]

        if "default" in kwargs:
            return super().get(name, ktype=ktype, **kwargs)
        else:
            # this will throw if name is not available
            return super()._get(name, ktype=ktype)

    def get_string(self, name):
        return self.get(name, ktype=str)

    def get_long(self, name):
        return self.get(name, ktype=int)

    def clone(self):
        return self._from_raw_handle(eccodes.codes_clone(self._handle))

    def set_multiple(self, values):
        try:
            assert self.path is None, "Only cloned handles can have values changed"
            eccodes.codes_set_key_vals(self._handle, values)
        except Exception as e:
            LOG.error(f"Error setting: {values}")
            LOG.exception(e)
            raise

    def set_long(self, name, value):
        try:
            assert self.path is None, "Only cloned handles can have values changed"
            eccodes.codes_set_long(self._handle, name, value)
        except Exception as e:
            LOG.error("Error setting %s=%s", name, value)
            LOG.exception(e)
            raise

    def set_double(self, name, value):
        try:
            assert self.path is None, "Only cloned handles can have values changed"
            eccodes.codes_set_double(self._handle, name, value)
        except Exception as e:
            LOG.error("Error setting %s=%s", name, value)
            LOG.exception(e)
            raise

    def set_string(self, name, value):
        try:
            assert self.path is None, "Only cloned handles can have values changed"
            eccodes.codes_set_string(self._handle, name, value)
        except Exception as e:
            LOG.error("Error setting %s=%s", name, value)
            LOG.exception(e)
            raise

    def set(self, name, value):
        try:
            assert self.path is None, "Only cloned handles can have values changed"

            if isinstance(value, list):
                return eccodes.codes_set_array(self._handle, name, value)

            return eccodes.codes_set(self._handle, name, value)
        except Exception as e:
            LOG.error("Error setting %s=%s", name, value)
            LOG.exception(e)
            raise

    def write(self, f):
        eccodes.codes_write(self._handle, f)

    def save(self, path):
        with open(path, "wb") as f:
            self.write_to(f)
            self.path = path
            self.offset = 0

    def read_bytes(self, offset, length):
        if self.path is not None:
            with open(self.path, "rb") as f:
                f.seek(offset)
                return f.read(length)


class ReaderLRUCache(dict):
    def __init__(self, size):
        self.readers = dict()
        self.lock = threading.Lock()
        self.size = size

    def __getitem__(self, path_and_cls):
        path = path_and_cls[0]
        cls = path_and_cls[1]
        key = (path, os.getpid())
        with self.lock:
            try:
                return super().__getitem__(key)
            except KeyError:
                pass

            c = self[key] = cls(path)
            while len(self) >= self.size:
                _, oldest = min((v.last, k) for k, v in self.items())
                del self[oldest]

            return c


cache = ReaderLRUCache(32)  # TODO: Add to config


class CodesReader:
    PRODUCT_ID = None
    HANDLE_TYPE = None

    def __init__(self, path):
        self.path = path
        self.lock = threading.Lock()
        # print("OPEN", self.path)
        self.file = open(self.path, "rb")
        self.last = time.time()

    def __del__(self):
        try:
            # print("CLOSE", self.path)
            self.file.close()
        except Exception:
            pass

    @classmethod
    def from_cache(cls, path):
        return cache[(path, cls)]

    def at_offset(self, offset):
        with self.lock:
            self.last = time.time()
            self.file.seek(offset, 0)
            handle = eccodes.codes_new_from_file(
                self.file,
                self.PRODUCT_ID,
            )
            assert handle is not None
            return self.HANDLE_TYPE(handle, self.path, offset)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.path}"



import typing as T

import numpy as np
import pandas as pd
import xarray as xr

dummy_attribute_1 = 1
dummy_attribute_2 = 2

XR_TYPES = (xr.Dataset, xr.DataArray, xr.Variable)


def numpy_mean(
    a: T.Union[
        np.ndarray, xr.Dataset, xr.DataArray, xr.Variable, pd.DataFrame, pd.Series
    ],
    **kwargs
):
    return np.mean(a, **kwargs)


def xarray_ones_like(dataarray: T.Union[xr.DataArray, xr.Dataset], **kwargs):
    return xr.ones_like(dataarray, **kwargs)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from earthkit.data import from_source

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from indexing_fixtures import get_tmp_fixture  # noqa


@pytest.mark.parametrize("mode", ["file", "multi", "directory"])
@pytest.mark.parametrize(
    "params,expected_meta,metadata_keys",
    [
        (dict(param=1, level=2), [["u", 500]], ["shortName", "level:l"]),
        (dict(param=1, level=2), [[131, 500]], ["paramId", "level:l"]),
        (
            dict(param=[0, 1], level=[2, 3]),
            [
                ["t", 500],
                ["t", 700],
                ["u", 500],
                ["u", 700],
            ],
            ["shortName", "level:l"],
        ),
        (
            dict(param=[0], level=[3, 2], type=0),
            [
                ["t", 500, "an"],
                ["t", 700, "an"],
            ],
            ["shortName", "level:l", "marsType"],
        ),
        (
            dict(level=-1),
            [
                ["t", 1000],
                ["u", 1000],
                ["v", 1000],
            ],
            ["shortName", "level:l"],
        ),
    ],
)
def test_indexing_isel_grib_file(mode, params, expected_meta, metadata_keys):
    tmp, path = get_tmp_fixture(mode)
    ds = from_source("file", path, indexing=True)
    assert len(ds) == 18

    g = ds.isel(**params)
    assert len(g) == len(expected_meta)

    # we sort the result to make the contents checking simpler
    g = g.order_by(["param", "level"])

    if len(expected_meta) > 0:
        keys = list(params.keys())
        if metadata_keys:
            keys = metadata_keys

        assert g.metadata(keys) == expected_meta


def test_indexing_isel_grib_file_invalid_key():
    tmp, path = get_tmp_fixture("file")
    ds = from_source("file", path, indexing=True)
    assert len(ds) == 18

    r = ds.isel(INVALIDKEY=0)
    assert len(r) == 0

    with pytest.raises(IndexError):
        ds.isel(level=500)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import os
import tempfile

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.readers.netcdf import NetCDFField
from earthkit.data.testing import (
    earthkit_examples_file,
    earthkit_file,
    earthkit_test_data_file,
)


def check_array(v, shape=None, first=None, last=None, meanv=None, eps=1e-3):
    assert v.shape == shape
    assert np.isclose(v[0], first, eps)
    assert np.isclose(v[-1], last, eps)
    assert np.isclose(v.mean(), meanv, eps)


def test_netcdf():
    for s in from_source("file", earthkit_file("docs/examples/test.nc")):
        s is not None


def test_dummy_netcdf_reader_1():
    ds = from_source("file", earthkit_file("docs/examples/test.nc"))
    # assert str(ds).startswith("NetCDFReader"), r
    assert len(ds) == 2
    assert isinstance(ds[1], NetCDFField), ds


@pytest.mark.parametrize("attribute", ["coordinates", "bounds", "grid_mapping"])
def test_dummy_netcdf_reader_2(attribute):
    s = from_source(
        "dummy-source",
        kind="netcdf",
        attributes={"a": {attribute: f"{attribute}_of_a"}},
        variables=["a", f"{attribute}_of_a"],
    )
    ds = s.to_xarray()
    assert "lat" in ds.dims
    assert len(s) == 1
    # s.to_datetime_list()
    s.bounding_box()


def test_dummy_netcdf():
    s = from_source("dummy-source", kind="netcdf")
    ds = s.to_xarray()
    assert "lat" in ds.dims


def test_dummy_netcdf_2():
    s = from_source(
        "dummy-source", kind="netcdf", dims=["lat", "lon", "time"], variables=["a", "b"]
    )
    ds = s.to_xarray()
    assert "lat" in ds.dims


def test_dummy_netcdf_3():
    s = from_source(
        "dummy-source",
        kind="netcdf",
        dims={"lat": dict(size=3), "lon": dict(size=2), "time": dict(size=2)},
        variables=["a", "b"],
    )
    ds = s.to_xarray()
    assert "lat" in ds.dims


def test_dummy_netcdf_4():
    s = from_source(
        "dummy-source",
        kind="netcdf",
        dims={"lat": dict(size=3), "lon": dict(size=2), "time": dict(size=2)},
        variables={
            "a": dict(dims=["lat", "lon"]),
            "b": dict(dims=["lat", "time"]),
        },
    )
    ds = s.to_xarray()
    assert "lat" in ds.dims


@pytest.mark.long_test
def test_netcdf_multi_cds():
    if not os.path.exists(os.path.expanduser("~/.cdsapirc")):
        pytest.skip("No ~/.cdsapirc")
    s1 = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        product_type="reanalysis",
        param="2t",
        date="2021-03-01",
        grid=[20, 20],
        format="netcdf",
    )
    s1.to_xarray()
    s2 = from_source(
        "cds",
        "reanalysis-era5-single-levels",
        product_type="reanalysis",
        param="2t",
        date="2021-03-02",
        grid=[20, 20],
        format="netcdf",
    )
    s2.to_xarray()

    source = from_source("multi", s1, s2)
    for s in source:
        print(s)

    source.to_xarray()


def test_netcdf_multi_sources():
    path = earthkit_test_data_file("era5_2t_1.nc")
    s1 = from_source("file", path)
    s1.to_xarray()
    assert s1.path == path

    path = earthkit_test_data_file("era5_2t_2.nc")
    s2 = from_source("file", path)
    s2.to_xarray()
    assert s2.path == path

    s3 = from_source("multi", s1, s2)
    for s in s3:
        print(s)

    assert len(s3) == 2
    assert s3[0].datetime() == {
        "base_time": datetime.datetime(2021, 3, 1, 12, 0),
        "valid_time": datetime.datetime(2021, 3, 1, 12, 0),
    }
    assert s3[1].datetime() == {
        "base_time": datetime.datetime(2021, 3, 2, 12, 0),
        "valid_time": datetime.datetime(2021, 3, 2, 12, 0),
    }
    assert s3.datetime() == {
        "base_time": [
            datetime.datetime(2021, 3, 1, 12, 0),
            datetime.datetime(2021, 3, 2, 12, 0),
        ],
        "valid_time": [
            datetime.datetime(2021, 3, 1, 12, 0),
            datetime.datetime(2021, 3, 2, 12, 0),
        ],
    }
    s3.to_xarray()


def test_netcdf_multi_files():
    ds = from_source(
        "file",
        [
            earthkit_test_data_file("era5_2t_1.nc"),
            earthkit_test_data_file("era5_2t_2.nc"),
        ],
    )

    assert len(ds) == 2
    assert ds[0].datetime() == {
        "base_time": datetime.datetime(2021, 3, 1, 12, 0),
        "valid_time": datetime.datetime(2021, 3, 1, 12, 0),
    }
    assert ds[1].datetime() == {
        "base_time": datetime.datetime(2021, 3, 2, 12, 0),
        "valid_time": datetime.datetime(2021, 3, 2, 12, 0),
    }
    assert ds.datetime() == {
        "base_time": [
            datetime.datetime(2021, 3, 1, 12, 0),
            datetime.datetime(2021, 3, 2, 12, 0),
        ],
        "valid_time": [
            datetime.datetime(2021, 3, 1, 12, 0),
            datetime.datetime(2021, 3, 2, 12, 0),
        ],
    }

    ds.to_xarray()


def test_get_fields_missing_standard_name_attr_in_coord_array():
    """test _get_fields() can handle a missing 'standard_name' attr in coordinate data arrays"""

    # example dataset
    fs = from_source("file", earthkit_examples_file("test.nc"))
    ds = fs.to_xarray()

    # delete 'standard_name' attribute (if exists) in any coordinate data arrays
    for coord_name in ds.coords:
        try:
            del ds.coords[coord_name].attrs["standard_name"]
        except Exception as e:
            print(e)

    # save updates to disk and try read that file source
    with tempfile.TemporaryDirectory() as tmp_dir:
        fpath = os.path.join(tmp_dir, "tmp.nc")
        ds.to_netcdf(fpath)
        fs = from_source("file", earthkit_test_data_file(fpath))
        assert len(fs) == 2


if __name__ == "__main__":
    from earthkit.data.testing import main

    # test_datetime()
    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import weakref
from importlib import import_module

from earthkit.data.core import Base
from earthkit.data.core.settings import SETTINGS
from earthkit.data.decorators import locked

LOG = logging.getLogger(__name__)


class ReaderMeta(type(Base), type(os.PathLike)):
    pass


class Reader(Base, os.PathLike, metaclass=ReaderMeta):
    appendable = False  # Set to True if the data can be appened to and existing file
    binary = True

    def __init__(self, source, path):
        LOG.debug("Reader for %s is %s", path, self.__class__.__name__)

        self._source = weakref.ref(source)
        self.path = path

    @property
    def source(self):
        return self._source()

    @property
    def filter(self):
        return self.source.filter

    @property
    def merger(self):
        return self.source.merger

    def mutate(self):
        # Give a chance to `directory` or `zip` to change the reader
        return self

    def mutate_source(self):
        # The source may ask if it needs to mutate
        return None

    def ignore(self):
        # Used by multi-source
        return False

    def cache_file(self, *args, **kwargs):
        return self.source.cache_file(*args, **kwargs)

    def save(self, path):
        mode = "wb" if self.binary else "w"
        with open(path, mode) as f:
            self.write(f)

    def write(self, f):
        if not self.appendable:
            assert f.tell() == 0
        mode = "rb" if self.binary else "r"
        with open(self.path, mode) as g:
            while True:
                chunk = g.read(1024 * 1024)
                if not chunk:
                    break
                f.write(chunk)

    def __fspath__(self):
        return self.path

    def index_content(self):
        LOG.warning(f"index-content(): Ignoring {self.path}")
        return []


_READERS = {}


# TODO: Add plugins
@locked
def _readers(method_name):
    if not _READERS:
        here = os.path.dirname(__file__)
        for path in sorted(os.listdir(here)):
            if path[0] in ("_", "."):
                continue

            if path.endswith(".py") or os.path.isdir(os.path.join(here, path)):
                name, _ = os.path.splitext(path)
                try:
                    for method in ["reader", "memory_reader", "stream_reader"]:
                        module = import_module(f".{name}", package=__name__)
                        if hasattr(module, method):
                            _READERS[(name, method)] = getattr(module, method)
                            if hasattr(module, "aliases"):
                                for a in module.aliases:
                                    assert a not in _READERS
                                    _READERS[(a, method_name)] = getattr(module, method)
                except Exception:
                    LOG.exception("Error loading reader %s", name)

    return {k[0]: v for k, v in _READERS.items() if k[1] == method_name}


def _find_reader(method_name, source, path_or_bufr_or_stream, magic):
    """Helper function to create a reader.

    Tries all the registered methods stored in _READERS.
    """
    for deeper_check in (False, True):
        # We do two passes, the second one
        # allow the plugin to look deeper in the buffer
        for name, r in _readers(method_name).items():
            reader = r(source, path_or_bufr_or_stream, magic, deeper_check)
            if reader is not None:
                return reader.mutate()

    from .unknown import Unknown

    return Unknown(
        source, path_or_bufr_or_stream if method_name == "reader" else "", magic
    )


def reader(source, path):
    """Create the reader for a file/directory specified by path"""
    assert isinstance(path, str), source

    if hasattr(source, "reader"):
        reader = source.reader
        LOG.debug("Looking for a reader for %s (%s)", path, reader)
        if callable(reader):
            return reader(source, path)
        if isinstance(reader, str):
            return _readers()[reader.replace("-", "_")](source, path, None, False)

        raise TypeError(
            "Provided reader must be a callable or a string, not %s" % type(reader)
        )

    if os.path.isdir(path):
        from .directory import DirectoryReader

        return DirectoryReader(source, path).mutate()
    LOG.debug("Reader for %s", path)

    n_bytes = SETTINGS.get("reader-type-check-bytes")
    with open(path, "rb") as f:
        magic = f.read(n_bytes)

    LOG.debug("Looking for a reader for %s (%s)", path, magic)

    return _find_reader("reader", source, path, magic)


def memory_reader(source, buf):
    """Create a reader for data held in a memory buffer"""
    assert isinstance(buf, (bytes, bytearray)), source
    n_bytes = SETTINGS.get("reader-type-check-bytes")
    magic = buf[: min(n_bytes, len(buf) - 1)]
    return _find_reader("memory_reader", source, buf, magic)


def stream_reader(source, stream):
    """Create a reader for a stream"""
    magic = None
    if hasattr(stream, "peek") and callable(stream.peek):
        try:
            n_bytes = SETTINGS.get("reader-type-check-bytes")
            magic = stream.peek(n_bytes)
            if len(magic) > n_bytes:
                magic = magic[:n_bytes]
        except Exception:
            pass

    return _find_reader("stream_reader", source, stream, magic)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import xarray as xr

from earthkit.data.translators import Translator


class XArrayDataArrayTranslator(Translator):
    """Wrapper around an xarray `DataArray`, offering polymorphism and
    convenience methods.
    """

    def __init__(self, data, *args, **kwargs):
        self.data = data.to_xarray(*args, **kwargs)

    def __call__(self):
        """
        Data-Array requested, if Dataset return the first data variable in dataset.
        """
        # import xarray as xr

        if isinstance(self.data, xr.Dataset):
            first_data_var = list(self.data.data_vars)[0]
            return self.data[first_data_var]

        return self.data


class XArrayDatasetTranslator(XArrayDataArrayTranslator):
    """
    Wrapper around an xarray `DataSet`, offering polymorphism and convenience
    methods.
    """

    def __call__(self):
        """
        Dataset requested, if DataArray convert to Dataset.
        """
        # import xarray as xr

        if isinstance(self.data, xr.DataArray):
            return self.data.to_dataset()

        return self.data


def translator(data, cls, *args, **kwargs):
    # import xarray as xr

    if cls in [xr.DataArray]:
        return XArrayDataArrayTranslator(data, *args, **kwargs)
    if cls in [xr.Dataset]:
        return XArrayDatasetTranslator(data, *args, **kwargs)

    return None



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import (
    earthkit_examples_file,
    earthkit_remote_test_data_file,
    earthkit_test_data_file,
)
from earthkit.data.utils import projections


def check_array(v, shape=None, first=None, last=None, meanv=None, eps=1e-3):
    assert v.shape == shape
    assert np.isclose(v[0], first, eps)
    assert np.isclose(v[-1], last, eps)
    assert np.isclose(v.mean(), meanv, eps)


def test_netcdf_to_points_1():
    ds = from_source("file", earthkit_test_data_file("test_single.nc"))

    eps = 1e-5
    v = ds[0].to_points(flatten=True)
    assert isinstance(v, dict)
    assert isinstance(v["x"], np.ndarray)
    assert isinstance(v["y"], np.ndarray)
    check_array(
        v["x"],
        (84,),
        first=0.0,
        last=330.0,
        meanv=165.0,
        eps=eps,
    )
    check_array(
        v["y"],
        (84,),
        first=90,
        last=-90,
        meanv=0,
        eps=eps,
    )


def test_bbox():
    ds = from_source("file", earthkit_examples_file("test.nc"))
    bb = ds.bounding_box()
    assert len(bb) == 2
    for b in bb:
        assert b.as_tuple() == (73, -27, 33, 45)


def test_netcdf_proj_string_non_cf():
    f = from_source("file", earthkit_examples_file("test.nc"))
    with pytest.raises(AttributeError):
        f[0].projection()


def test_netcdf_projection_laea():
    f = from_source("url", earthkit_remote_test_data_file("examples", "efas.nc"))
    projection = f[0].projection()
    assert isinstance(projection, projections.LambertAzimuthalEqualArea)
    assert projection.parameters == {
        "central_latitude": 52.0,
        "central_longitude": 10.0,
        "false_northing": 3210000.0,
        "false_easting": 4321000.0,
    }
    assert projection.globe == {
        "ellipse": "GRS80",
    }


def test_netcdf_proj_string_laea():
    f = from_source("url", earthkit_remote_test_data_file("examples", "efas.nc"))
    r = f[0].projection()
    assert (
        r.to_proj_string()
        == "+proj=laea +lat_0=52 +lon_0=10 +x_0=4321000 +y_0=3210000 +ellps=GRS80 +units=m +no_defs"
    )


if __name__ == "__main__":
    from earthkit.data.testing import main

    # test_datetime()
    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
from contextlib import closing
from itertools import product

import numpy as np

from earthkit.data.core.fieldlist import Field, FieldList
from earthkit.data.core.geography import Geography
from earthkit.data.core.index import Index, MaskIndex, MultiIndex
from earthkit.data.core.metadata import RawMetadata
from earthkit.data.utils.bbox import BoundingBox
from earthkit.data.utils.dates import to_datetime
from earthkit.data.utils.projections import Projection

from . import Reader

GEOGRAPHIC_COORDS = {
    "x": ["x", "projection_x_coordinate", "lon", "longitude"],
    "y": ["y", "projection_y_coordinate", "lat", "latitude"],
}


def as_datetime(self, time):
    return datetime.datetime.strptime(str(time)[:19], "%Y-%m-%dT%H:%M:%S")


def as_level(self, level):
    n = float(level)
    if int(n) == n:
        return int(n)
    return n


class Slice:
    def __init__(self, name, value, index, is_dimension, is_info):
        self.name = name
        self.index = index
        self.value = value
        self.is_dimension = (not is_info,)
        self.is_info = is_info

    def __repr__(self):
        return "[%s:%s=%s]" % (self.name, self.index, self.value)


class TimeSlice(Slice):
    pass


class Coordinate:
    def __init__(self, variable, info):
        self.variable = variable
        # We only support 1D coordinate for now
        # assert len(variable.dims) == 1
        self.is_info = info
        self.is_dimension = not info

        if variable.values.ndim == 0:
            self.values = [self.convert(variable.values)]
        else:
            self.values = [self.convert(t) for t in variable.values.flatten()]

    def make_slice(self, value):
        return self.slice_class(
            self.variable.name,
            value,
            self.values.index(value),
            self.is_dimension,
            self.is_info,
        )

    def __repr__(self):
        return "%s[name=%s,values=%s]" % (
            self.__class__.__name__,
            self.variable.name,
            len(self.values),
        )


class TimeCoordinate(Coordinate):
    slice_class = TimeSlice
    is_dimension = True
    convert = as_datetime


class LevelCoordinate(Coordinate):
    # This class is just in case we want to specialise
    # 'level', othewise, it is the same as OtherCoordinate
    slice_class = Slice
    is_dimension = False
    convert = as_level


class OtherCoordinate(Coordinate):
    slice_class = Slice
    is_dimension = False
    convert = as_level


class DataSet:
    def __init__(self, ds):
        self._ds = ds
        self._bbox = {}
        self._cache = {}

    @property
    def data_vars(self):
        return self._ds.data_vars

    def __getitem__(self, key):
        if key not in self._cache:
            self._cache[key] = self._ds[key]
        return self._cache[key]

    def bbox(self, variable):
        data_array = self[variable]
        dims = data_array.dims

        lat = dims[-2]
        lon = dims[-1]

        if (lat, lon) not in self._bbox:
            dims = data_array.dims

            latitude = data_array[lat]
            longitude = data_array[lon]

            self._bbox[(lat, lon)] = (
                np.amax(latitude.data),
                np.amin(longitude.data),
                np.amin(latitude.data),
                np.amax(longitude.data),
            )

        return self._bbox[(lat, lon)]


class NetCDFFieldGeography(Geography):
    def __init__(self, metadata, da, ds, variable):
        self.metadata = metadata
        self._da = da
        self._ds = ds
        self.north, self.west, self.south, self.east = self._ds.bbox(variable)

    def latitudes(self):
        return self.x()

    def longitudes(self):
        return self.y()

    def _get_xy(self, axis, flatten=False):
        if axis not in ("x", "y"):
            raise ValueError(f"Invalid axis={axis}")

        points = dict()
        for ax in ("x", "y"):
            for coord in self._da.coords:
                if self._da.coords[coord].attrs.get("axis", "").lower() == ax:
                    break
            else:
                candidates = GEOGRAPHIC_COORDS.get(ax, [])
                for coord in candidates:
                    if coord in self._da.coords:
                        break
                else:
                    raise ValueError(f"No coordinate found with axis '{ax}'")
            points[ax] = self._da.coords[coord]
        points["x"], points["y"] = np.meshgrid(points["x"], points["y"])
        if flatten:
            points[axis] = points[axis].flatten()
        return points[axis]

    def x(self):
        return self._get_xy("x", flatten=True)

    def y(self):
        return self._get_xy("y", flatten=True)

    def shape(self):
        return self._da.shape[-2:]

    def _unique_grid_id(self):
        return self.shape

    def projection(self):
        return Projection.from_cf_grid_mapping(**self._grid_mapping().attrs)

    def bounding_box(self):
        return BoundingBox(
            north=self.north, south=self.south, east=self.east, west=self.west
        )

    def _grid_mapping(self):
        if "grid_mapping" in self._da.attrs:
            grid_mapping = self._ds[self._da.attrs["grid_mapping"]]
        else:
            raise AttributeError(
                "no CF-compliant 'grid_mapping' detected in netCDF attributes"
            )
        return grid_mapping


class NetCDFMetadata(RawMetadata):
    def __init__(self, field):
        if not isinstance(field, NetCDFField):
            raise TypeError(
                f"NetCDFMetadata: expected field type NetCDFField, got {type(field)}"
            )
        self._field = field
        self._geo = None

        d = dict(self._field._da.attrs)
        d["variable"] = self._field.variable
        for s in self._field.slices:
            if isinstance(s, TimeSlice):
                d[s.name] = to_datetime(s.value)
            else:
                d[s.name] = s.value
        super().__init__(d)

    def override(self, *args, **kwargs):
        return None

    @property
    def geography(self):
        if self._geo is None:
            self._geo = NetCDFFieldGeography(
                self, self._field._da, self._field._ds, self._field.variable
            )
        return self._geo

    def datetime(self):
        t = to_datetime(self._field.time)
        return {"base_time": t, "valid_time": t}

    def ls_keys(self):
        LS_KEYS = ["variable", "level", "time", "units"]
        return LS_KEYS


class NetCDFField(Field):
    def __init__(self, ds, variable, slices, non_dim_coords):
        super().__init__()

        self._ds = ds
        self._da = ds[variable]

        # self.north, self.west, self.south, self.east = ds.bbox(variable)

        self.variable = variable
        self.slices = slices
        self.non_dim_coords = non_dim_coords
        # self.name = self.variable

        # print(f"ds={ds}")
        # print(f"da={data_array}")
        # print(f"non_dim_coords={non_dim_coords}")

        self.title = getattr(
            self._da,
            "long_name",
            getattr(self._da, "standard_name", self.variable),
        )

        self.time = non_dim_coords.get("valid_time", non_dim_coords.get("time"))

        # print('====', non_dim_coords)

        # print(f"time={self.time}")

        for s in self.slices:
            if isinstance(s, TimeSlice):
                self.time = s.value

            if s.is_info:
                self.title += " (" + s.name + "=" + str(s.value) + ")"

        # print(f"-> time={self.time}")

    def __repr__(self):
        return (
            f"NetCDFField({self.variable},"
            + ",".join([f"{s.name}={s.value}" for s in self.slices])
            + ")"
        )

    def _make_metadata(self):
        return NetCDFMetadata(self)

    def to_xarray(self):
        dims = self._da.dims
        v = {}
        for s in self.slices:
            if s.is_dimension:
                if s.name in dims:
                    v[s.name] = s.index
        return self._da.isel(**v)

    def to_pandas(self):
        return self.to_xarray().to_pandas()

    def _to_numpy(self):
        return self.to_xarray().to_numpy()

    @property
    def values(self):
        return self._to_numpy().flatten()

    def to_numpy(self, flatten=False, dtype=None):
        values = self._to_numpy()
        if not flatten:
            values = values.reshape(self.shape)
        else:
            values = values.flatten()
        if dtype is not None:
            values = values.astype(dtype)

        return values


class NetCDFFieldList(FieldList):
    def __init__(self, path, *args, **kwargs):
        self.path = path
        self._fields = None
        Index.__init__(self, *args, **kwargs)

    @property
    def fields(self):
        if self._fields is None:
            self._scan()
        return self._fields

    def _scan(self):
        if self._fields is None:
            self._fields = self._get_fields()

    def _get_fields(self):
        import xarray as xr

        with closing(
            xr.open_mfdataset(self.path, combine="by_coords")
        ) as ds:  # or nested
            return self._get_fields_from_ds(DataSet(ds))

    def _get_fields_from_ds(self, ds):  # noqa C901
        # Select only geographical variables
        has_lat = False
        has_lon = False

        fields = []

        skip = set()

        for name in ds.data_vars:
            v = ds[name]
            skip.update(getattr(v, "coordinates", "").split(" "))
            skip.update(getattr(v, "bounds", "").split(" "))
            skip.update(getattr(v, "grid_mapping", "").split(" "))

        for name in ds.data_vars:
            if name in skip:
                continue

            v = ds[name]

            coordinates = []

            # self.log.info('Scanning file: %s var=%s coords=%s', self.path, name, v.coords)

            info = [value for value in v.coords if value not in v.dims]
            non_dim_coords = {}
            for coord in v.coords:
                if coord not in v.dims:
                    non_dim_coords[coord] = ds[coord].values
                    continue

                c = ds[coord]

                # self.log.info("COORD %s %s %s %s", coord, type(coord), hasattr(c, 'calendar'), c)

                standard_name = getattr(c, "standard_name", "")
                axis = getattr(c, "axis", "")
                long_name = getattr(c, "long_name", "")

                use = False

                if (
                    standard_name.lower() in GEOGRAPHIC_COORDS["x"]
                    or (long_name == "longitude")
                    or (axis == "X")
                ):
                    has_lon = True
                    use = True

                if (
                    standard_name.lower() in GEOGRAPHIC_COORDS["y"]
                    or (long_name == "latitude")
                    or (axis == "Y")
                ):
                    has_lat = True
                    use = True

                # print(f"  standard_name={standard_name}")

                # Of course, not every one sets the standard_name
                if (
                    standard_name in ["time", "forecast_reference_time"]
                    or long_name in ["time"]
                    or axis == "T"
                ):
                    # we might not be able to convert time to datetime
                    try:
                        coordinates.append(TimeCoordinate(c, coord in info))
                        use = True
                    except ValueError:
                        break

                # TODO: Support other level types
                if standard_name in [
                    "air_pressure",
                    "model_level_number",
                    "altitude",
                ] or long_name in [
                    "pressure_level"
                ]:  # or axis == 'Z':
                    coordinates.append(LevelCoordinate(c, coord in info))
                    use = True

                if axis in ("X", "Y"):
                    use = True

                if not use:
                    coordinates.append(OtherCoordinate(c, coord in info))

            if not (has_lat and has_lon):
                # self.log.info("NetCDFReader: skip %s (Not a 2 field)", name)
                continue

            for values in product(*[c.values for c in coordinates]):
                slices = []
                for value, coordinate in zip(values, coordinates):
                    slices.append(coordinate.make_slice(value))

                fields.append(NetCDFField(ds, name, slices, non_dim_coords))

        # if not fields:
        #     raise Exception("NetCDFReader no 2D fields found in %s" % (self.path,))

        return fields

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(_, NetCDFFieldList) for _ in sources)
        return NetCDFMultiFieldList(sources)

    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return NetCDFMaskFieldList(*args, **kwargs)

    def to_pandas(self):
        return self.to_xarray().to_pandas()

    def to_xarray(self, **kwargs):
        return type(self).to_xarray_multi_from_paths(self.path, **kwargs)

    @classmethod
    def to_xarray_multi_from_paths(cls, paths, **kwargs):
        import xarray as xr

        if not isinstance(paths, list):
            paths = [paths]

        options = dict()
        options.update(kwargs.get("xarray_open_mfdataset_kwargs", {}))

        return xr.open_mfdataset(
            paths,
            **options,
        )


class NetCDFFieldListInFiles(NetCDFFieldList):
    pass


class NetCDFFieldListInOneFile(NetCDFFieldListInFiles):
    VERSION = 1

    def __init__(self, path, **kwargs):
        assert isinstance(path, str), path
        super().__init__(path, **kwargs)

    def __getitem__(self, n):
        if isinstance(n, int):
            return self.fields[n]
        else:
            return super().__getitem__(n)

    def __len__(self):
        return len(self.fields)


class NetCDFMaskFieldList(NetCDFFieldList, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)


class NetCDFMultiFieldList(NetCDFFieldList, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)

    def to_xarray(self, **kwargs):
        return NetCDFFieldList.to_xarray_multi_from_paths(
            [x.path for x in self.indexes], **kwargs
        )


class NetCDFFieldListReader(NetCDFFieldListInOneFile, Reader):
    def __init__(self, source, path):
        Reader.__init__(self, source, path)
        NetCDFFieldList.__init__(self, path)

    def __repr__(self):
        return "NetCDFFieldListReader(%s)" % (self.path,)

    # @classmethod
    # def merge(cls, readers):
    #     assert all(isinstance(s, NetCDFFieldListReader) for s in readers), readers
    #     assert len(readers) > 1

    #     return NetCDFFieldListReader(readers[0], [s.path for s in readers])

    def mutate_source(self):
        # A NetCDFReader is a source itself
        return self

    # def to_pandas(self):
    #     return self.to_xarray().to_pandas()

    # def to_xarray(self, **kwargs):
    #     return type(self).to_xarray_multi_from_paths(self.path, **kwargs)

    # @classmethod
    # def to_xarray_multi_from_paths(cls, paths, **kwargs):
    #     import xarray as xr

    #     if not isinstance(paths, list):
    #         paths = [paths]

    #     options = dict()
    #     options.update(kwargs.get("xarray_open_mfdataset_kwargs", {}))

    #     return xr.open_mfdataset(
    #         paths,
    #         **options,
    #     )


class NetCDFReader(Reader):
    def __init__(self, source, path):
        Reader.__init__(self, source, path)
        NetCDFFieldList.__init__(self, path)

    def __repr__(self):
        return "NetCDFReader(%s)" % (self.path,)

    def to_numpy(self, flatten=False):
        arr = self.to_xarray().to_array().to_numpy()
        if flatten:
            arr = arr.flatten()
        return arr

    def to_pandas(self):
        return self.to_xarray().to_pandas()

    def to_xarray(self, **kwargs):
        return type(self).to_xarray_multi_from_paths([self.path], **kwargs)

    @classmethod
    def to_xarray_multi_from_paths(cls, paths, **kwargs):
        import xarray as xr

        if not isinstance(paths, list):
            paths = [paths]

        options = dict()
        options.update(kwargs.get("xarray_open_mfdataset_kwargs", {}))

        return xr.open_mfdataset(
            paths,
            **options,
        )


def _match_magic(magic, deeper_check):
    if magic is not None:
        type_id = (b"\x89HDF", b"CDF\x01", b"CDF\x02")
        return len(magic) >= 4 and magic[:4] in type_id
    return False


def reader(source, path, magic=None, deeper_check=False):
    if _match_magic(magic, deeper_check):
        r = NetCDFFieldListReader(source, path)
        if len(r) > 0:
            return r
        else:
            return NetCDFReader(source, path)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os

import numpy as np

from earthkit.data import from_source
from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import earthkit_examples_file

LOG = logging.getLogger(__name__)


def test_numpy_list_grib_single_field():
    ds = from_source("file", earthkit_examples_file("test.grib"))

    assert ds[0].metadata("shortName") == "2t"

    v = ds[0].values
    v1 = v + 1

    md = ds[0].metadata()
    md1 = md.override(shortName="msl")
    r = FieldList.from_numpy(v1, md1)

    assert len(r) == 1
    assert np.allclose(v1, r[0].values)
    assert r[0].shape == ds[0].shape
    assert r[0].metadata("shortName") == "msl"

    # save to disk
    tmp = temp_file()
    r.save(tmp.path)
    assert os.path.exists(tmp.path)
    r_tmp = from_source("file", tmp.path)
    assert len(r_tmp) == 1
    assert np.allclose(v1, r_tmp[0].values)
    assert r_tmp[0].shape == ds[0].shape
    assert r_tmp[0].metadata("shortName") == "msl"


def test_numpy_list_grib_multi_field():
    ds = from_source("file", earthkit_examples_file("test.grib"))

    assert ds[0].metadata("shortName") == "2t"

    v = ds.values
    v1 = v + 1

    md1 = [f.metadata().override(shortName="2d") for f in ds]
    r = FieldList.from_numpy(v1, md1)

    assert len(r) == 2
    assert np.allclose(v1, r.values)
    for i, f in enumerate(r):
        assert f.shape == ds[i].shape
        assert f.metadata("shortName") == "2d", f"shortName {i}"
        assert f.metadata("name") == "2 metre dewpoint temperature", f"name {i}"

    # save to disk
    tmp = temp_file()
    r.save(tmp.path)
    assert os.path.exists(tmp.path)
    r_tmp = from_source("file", tmp.path)
    assert len(r_tmp) == 2
    assert np.allclose(v1, r_tmp.values)
    for i, f in enumerate(r_tmp):
        assert f.shape == ds[i].shape
        assert f.metadata("shortName") == "2d", f"shortName {i}"
        assert f.metadata("name") == "2 metre dewpoint temperature", f"name {i}"


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from earthkit.data.readers import stream_reader
from earthkit.data.sources.memory import MemoryBaseSource

from . import Source

LOG = logging.getLogger(__name__)


class StreamMemorySource(MemoryBaseSource):
    def __init__(self, reader, **kwargs):
        super().__init__(**kwargs)
        self._reader = reader

    def __iter__(self):
        return iter(self._reader)


class StreamSource(Source):
    def __init__(self, stream, group_by=None, **kwargs):
        super().__init__()
        self._stream = stream
        self._reader_ = None
        self._group_by = group_by if group_by is not None else []

        if isinstance(self._group_by, str):
            self._group_by = [self._group_by]
        else:
            try:
                self._group_by = list(self._group_by)
            except Exception:
                raise TypeError(f"unsupported types in `group_by`={self._group_by}")

        if self._group_by and not all([isinstance(x, str) for x in self._group_by]):
            raise TypeError(f"`group_by`={self._group_by} must contain str values")

        if self._group_by and "batch_size" in kwargs:
            raise TypeError(
                "got an invalid keyword argument. `batch_size` cannot be used when `group_by` is set"
            )

        self._batch_size = kwargs.pop("batch_size", 1)

        if self._batch_size < 0:
            raise ValueError(f"`batch_size`={self._batch_size} cannot be negative")
        if kwargs:
            raise TypeError(f"got invalid keyword argument(s): {list(kwargs.keys())}")

    @property
    def batch_size(self):
        return self._batch_size

    @property
    def group_by(self):
        return self._group_by

    def mutate(self):
        if self.batch_size == 0:
            source = StreamMemorySource(self._reader)
            return source
        else:
            return self

    def __iter__(self):
        return self

    def __next__(self):
        assert self.batch_size > 0
        if self.group_by:
            return self._reader.read_group(self.group_by)
        else:
            if self.batch_size == 1:
                return self._reader.__next__()
            else:
                return self._reader.read_batch(self.batch_size)

    @property
    def _reader(self):
        if self._reader_ is None:
            self._reader_ = stream_reader(self, self._stream)
            self._stream = None
        return self._reader_


source = StreamSource



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import pytest

from earthkit.data import from_source
from earthkit.data.core.metadata import RawMetadata
from earthkit.data.readers.grib.metadata import GribMetadata
from earthkit.data.testing import earthkit_examples_file, earthkit_test_data_file


@pytest.mark.parametrize(
    "params",
    [
        {"shortName": "2t", "perturbationNumber": 5},
        [("shortName", "2t"), ("perturbationNumber", 5)],
        RawMetadata({"shortName": "2t", "perturbationNumber": 5}),
    ],
)
def test_raw_metadata_create(params):
    md = RawMetadata(params)
    assert md["shortName"] == "2t"
    assert md["perturbationNumber"] == 5


def test_raw_metadata_create_with_kwarg():
    md = RawMetadata(shortName="2t", perturbationNumber=5)
    assert md["shortName"] == "2t"
    assert md["perturbationNumber"] == 5


def test_raw_metadata_get():
    md = RawMetadata({"shortName": "2t", "perturbationNumber": 5})

    assert len(md) == 2
    assert list(md.keys()) == ["shortName", "perturbationNumber"]
    assert [k for k in md] == ["shortName", "perturbationNumber"]
    assert {k: v for k, v in md.items()} == {"shortName": "2t", "perturbationNumber": 5}

    assert "shortName" in md
    assert "nonExistentKey" not in md

    assert md["shortName"] == "2t"
    assert md["perturbationNumber"] == 5
    assert md.get("shortName") == "2t"
    with pytest.raises(KeyError):
        md["nonExistentKey"]

    assert md.get("nonExistentKey") is None
    assert md.get("nonExistentKey", None) is None
    assert md.get("nonExistentKey", 12) == 12
    with pytest.raises(TypeError):
        md.get("centre", "shortName", "step")


@pytest.mark.parametrize(
    "params",
    [
        {"centre": "ecmf", "perturbationNumber": 8},
        [("centre", "ecmf"), ("perturbationNumber", 8)],
        RawMetadata({"centre": "ecmf", "perturbationNumber": 8}),
    ],
)
def test_raw_metadata_override(params):
    md = RawMetadata({"shortName": "2t", "perturbationNumber": 5})
    assert md["shortName"] == "2t"
    assert md["perturbationNumber"] == 5
    assert md.get("centre", None) is None

    md2 = md.override(params)
    assert id(md2) != id(md)
    assert md["shortName"] == "2t"
    assert md["perturbationNumber"] == 5
    assert md.get("centre", None) is None
    assert md2["shortName"] == "2t"
    assert md2["perturbationNumber"] == 8
    assert md2["centre"] == "ecmf"


def test_raw_metadata_override_with_kwarg():
    md = RawMetadata({"shortName": "2t", "perturbationNumber": 5})
    assert md["shortName"] == "2t"
    assert md["perturbationNumber"] == 5
    assert md.get("centre", None) is None

    md2 = md.override(centre="ecmf", perturbationNumber=8)
    assert id(md2) != id(md)
    assert md["shortName"] == "2t"
    assert md["perturbationNumber"] == 5
    assert md.get("centre", None) is None
    assert md2["shortName"] == "2t"
    assert md2["perturbationNumber"] == 8
    assert md2["centre"] == "ecmf"


def test_grib_metadata_create():
    f = from_source("file", earthkit_examples_file("test.grib"))
    md = f[0].metadata()
    assert isinstance(md, GribMetadata)

    # cannot create from dict
    with pytest.raises(TypeError):
        GribMetadata({"shortName": "u", "typeOfLevel": "pl", "levelist": 1000})

    # cannot create from raw metadata
    raw_md = RawMetadata({"shortName": "u", "typeOfLevel": "pl", "levelist": 1000})
    with pytest.raises(TypeError):
        GribMetadata(raw_md)


def test_grib_metadata_get():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    md = ds[0].metadata()

    assert len(md) == 192
    keys = list(md.keys())
    assert len(keys) == 192
    assert keys[0] == "globalDomain"
    assert keys[10] == "wrongPadding"

    keys = [k for k in md]
    assert len(keys) == 192
    assert keys[0] == "globalDomain"
    assert keys[10] == "wrongPadding"

    items = {k: v for k, v in md.items()}
    assert len(items) == 192
    assert items["shortName"] == "2t"
    assert items["typeOfLevel"] == "surface"

    assert "shortName" in md
    assert "nonExistentKey" not in md

    assert md["shortName"] == "2t"
    assert md["typeOfLevel"] == "surface"
    assert md.get("shortName") == "2t"

    with pytest.raises(KeyError):
        md["nonExistentKey"]

    assert md.get("nonExistentKey") is None
    assert md.get("nonExistentKey", None) is None
    assert md.get("nonExistentKey", 12) == 12

    sentinel = object()
    assert md.get("nonExistentKey", sentinel) is sentinel

    with pytest.raises(TypeError):
        md.get("centre", "shortName", "step")


def test_grib_grib_metadata_valid_datetime():
    ds = from_source("file", earthkit_test_data_file("t_time_series.grib"))
    md = ds[4].metadata()

    assert md["valid_datetime"] == datetime.datetime(2020, 12, 21, 18)


def test_grib_metadata_override():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    md = ds[0].metadata()

    md2 = md.override({"perturbationNumber": 5})
    assert id(md2) != id(md)
    assert md["perturbationNumber"] == 0
    assert md2["perturbationNumber"] == 5

    md2 = md.override({"shortName": "2d"})
    assert md["shortName"] == "2t"
    assert md2["shortName"] == "2d"

    md2 = md.override({"perturbationNumber": 5, "shortName": "2d"})
    assert md["perturbationNumber"] == 0
    assert md["shortName"] == "2t"
    assert md2["perturbationNumber"] == 5
    assert md2["shortName"] == "2d"

    md3 = md2.override({"step": 24})
    assert md["step"] == 0
    assert md2["step"] == 0
    assert md3["step"] == 24

    # all the handles should exist and be different
    assert md._handle._handle is not None
    assert md2._handle._handle is not None
    assert md3._handle._handle is not None
    assert md._handle._handle != md2._handle._handle
    assert md2._handle._handle != md3._handle._handle
    assert md._handle._handle != md3._handle._handle

    md = None
    assert md2._handle._handle is not None
    assert md3._handle._handle is not None


@pytest.mark.parametrize(
    "params",
    [
        {"perturbationNumber": 5, "shortName": "2d"},
        [("perturbationNumber", 5), ("shortName", "2d")],
        RawMetadata({"perturbationNumber": 5, "shortName": "2d"}),
    ],
)
def test_grib_metadata_override_1(params):
    ds = from_source("file", earthkit_examples_file("test.grib"))
    md = ds[0].metadata()
    assert md["perturbationNumber"] == 0
    assert md["shortName"] == "2t"

    md2 = md.override(params)
    assert id(md2) != id(md)
    assert md["perturbationNumber"] == 0
    assert md["shortName"] == "2t"
    assert md2["perturbationNumber"] == 5
    assert md2["shortName"] == "2d"


def test_grib_metadata_override_with_kwarg():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    md = ds[0].metadata()
    assert md["perturbationNumber"] == 0
    assert md["shortName"] == "2t"

    md2 = md.override(perturbationNumber=5, shortName="2d")
    assert id(md2) != id(md)
    assert md["perturbationNumber"] == 0
    assert md["shortName"] == "2t"
    assert md2["perturbationNumber"] == 5
    assert md2["shortName"] == "2d"


def test_grib_metadata_override_invalid():
    ds = from_source("file", earthkit_examples_file("test.grib"))
    md = ds[0].metadata()

    # invalid key
    with pytest.raises(Exception) as e:
        md.override({"__invalidkey": 5})
    assert "KeyValueNotFoundError" in e.typename

    # invalid value
    with pytest.raises(Exception) as e:
        md.override({"level": -100})
    assert "EncodingError" in e.typename


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import numpy as np

from earthkit.data.core.fieldlist import Field, FieldList
from earthkit.data.core.index import MaskIndex, MultiIndex
from earthkit.data.core.metadata import Metadata

LOG = logging.getLogger(__name__)


class NumpyField(Field):
    def __init__(self, array, metadata):
        self._array = array
        super().__init__(metadata=metadata)

    def _make_metadata(self):
        pass

    @property
    def values(self):
        return self._array

    def write(self, f):
        from earthkit.data.writers import write

        write(f, self.values, self._metadata, check_nans=False)


class NumpyFieldList(FieldList):
    def __init__(self, array, metadata, *args, **kwargs):
        self._array = array
        self._metadata = metadata

        if not isinstance(self._metadata, list):
            self._metadata = [self._metadata]

        for md in self._metadata:
            if not isinstance(md, Metadata):
                raise TypeError("metadata must be a subclass of MetaData")

        if self._array.shape[0] != len(self._metadata):
            import numpy as np

            # we have a single array and a single metadata
            if len(self._metadata) == 1 and self._shape_match(
                self._array.shape, self._metadata[0].geography.shape()
            ):
                self._array = np.array([self._array])
            else:
                raise ValueError(
                    (
                        f"first array dimension ({self._array.shape[0]}) differs "
                        f"from number of metadata objects ({len(self._metadata)})"
                    )
                )

        super().__init__(*args, **kwargs)

    def __getitem__(self, n):
        return NumpyField(self._array[n], self._metadata[n])

    def __len__(self):
        return self._array.shape[0]

    def _shape_match(self, shape1, shape2):
        if shape1 == shape2:
            return True
        if len(shape1) == 1 and shape1[0] == np.prod(shape2):
            return True
        return False


class NumpyMaskFieldList(NumpyFieldList, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)


class NumpyMultiFieldList(NumpyFieldList, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging

from . import Database

LOG = logging.getLogger(__name__)


class StdoutDatabase(Database):
    def __init__(self, db_path):
        pass

    def load(self, iterator):
        count = 0
        for entry in iterator:
            count += 1
            print(json.dumps(entry))
        return count



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging

from earthkit.data.core.fieldlist import Field
from earthkit.data.core.geography import Geography
from earthkit.data.core.metadata import RawMetadata
from earthkit.data.readers.grib.index import GribFieldList
from earthkit.data.readers.grib.metadata import GribMetadata
from earthkit.data.utils.bbox import BoundingBox
from earthkit.data.utils.projections import Projection

LOG = logging.getLogger(__name__)


class VirtualGribGeography(Geography):
    pass


class VirtualGribMetadata(RawMetadata):
    KEY_TYPES = {
        "s": str,
        "l": int,
        "d": float,
        "str": str,
        "int": int,
        "float": float,
        "": None,
    }

    KEY_TYPE_SUFFIX = {str: ":s", int: ":l", float: ":d"}

    KEY_GROUPS = {
        ("dataDate", "date"),
        ("dataTime", "time"),
        ("level", "levelist"),
        ("step", "endStep", "stepRange"),
    }

    def __init__(self, m):
        super().__init__(m)

    def _get_internal_key(self, key, astype=None, default=None, raise_on_missing=True):
        def _key_name(key):
            if key == "param":
                key = "shortName"
            elif key == "_param_id":
                key = "paramId"
            return key

        key = _key_name(key)
        key, _, key_type_str = key.partition(":")
        try:
            key_type = self.KEY_TYPES[key_type_str]
        except KeyError:
            raise ValueError(f"Key type={key_type_str} not supported")

        if key not in self:
            for v in self.KEY_GROUPS:
                if key in v:
                    for k in v:
                        if k in self:
                            key = k
                            break

        if key_type is None:
            key_type = astype

        if key == "stepRange" and key_type is None:
            key_type = str

        return super()._get_internal_key(
            key, astype=key_type, default=default, raise_on_missing=raise_on_missing
        )

    def shape(self):
        Nj = self.get("Nj", None)
        Ni = self.get("Ni", None)
        if Ni is None or Nj is None:
            n = len(self.get("values"))
            return (n,)  # shape must be a tuple
        return (Nj, Ni)

    def as_namespace(self, ns):
        return {}

    def ls_keys(self):
        return GribMetadata.LS_KEYS

    def namespaces(self):
        return []

    def latitudes(self):
        return self.get("latitudes")

    def longitudes(self):
        return self.get("longitudes")

    def x(self):
        grid_type = self.get("gridType", None)
        if grid_type in ["regular_ll", "reduced_gg", "regular_gg"]:
            return self.longitudes()

    def y(self):
        grid_type = self.get("gridType", None)
        if grid_type in ["regular_ll", "reduced_gg", "regular_gg"]:
            return self.latitudes()

    def _unique_grid_id(self):
        return self.get("md5GridSection", None)

    def datetime(self):
        return {
            "base_time": self._base_datetime(),
            "valid_time": self._valid_datetime(),
        }

    def _base_datetime(self):
        date = self.get("date", None)
        time = self.get("time", None)
        return datetime.datetime(
            date // 10000,
            date % 10000 // 100,
            date % 100,
            time // 100,
            time % 100,
        )

    def _valid_datetime(self):
        step = self.get("endStep", None)
        return self._base_datetime() + datetime.timedelta(hours=step)

    def projection(self):
        return Projection.from_proj_string(self.get("projTargetString", None))

    def bounding_box(self):
        return BoundingBox(
            north=self.get("latitudeOfFirstGridPointInDegrees", None),
            south=self.get("latitudeOfLastGridPointInDegrees", None),
            west=self.get("longitudeOfFirstGridPointInDegrees", None),
            east=self.get("longitudeOfLastGridPointInDegrees", None),
        )

    from earthkit.data.core.geography import Geography


class VirtualGribField(Field):
    def __init__(self, d):
        super().__init__(metadata=VirtualGribMetadata(d))

    @property
    def values(self):
        return self._metadata["values"]

    def _make_metadata(self):
        pass


class GribFromDicts(GribFieldList):
    def __init__(self, list_of_dicts, *args, **kwargs):
        self.list_of_dicts = list_of_dicts
        super().__init__(*args, **kwargs)

    def __getitem__(self, n):
        return VirtualGribField(self.list_of_dicts[n])

    def __len__(self):
        return len(self.list_of_dicts)


source = GribFromDicts



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
from abc import abstractmethod

import eccodes
from pdbufr.high_level_bufr.bufr import bufr_code_is_coord

from earthkit.data.core import Base
from earthkit.data.core.index import Index, MaskIndex, MultiIndex
from earthkit.data.utils.message import (
    CodesHandle,
    CodesMessagePositionIndex,
    CodesReader,
)
from earthkit.data.utils.metadata import metadata_argument
from earthkit.data.utils.parts import Part
from earthkit.data.utils.summary import make_bufr_html_tree

from .. import Reader
from .pandas import PandasMixIn

BUFR_LS_KEYS = {
    "edition": "edition",
    "type": "dataCategory",
    "subtype": "dataSubCategory",
    "c": "bufrHeaderCentre",
    "mv": "masterTablesVersionNumber",
    "lv": "localTablesVersionNumber",
    "subsets": "numberOfSubsets",
    "compr": "compressedData",
    "typicalDate": "typicalDate",
    "typicalTime": "typicalTime",
    "ident": "ident",
    "lat": "localLatitude",
    "lon": "localLongitude",
}


class BufrCodesMessagePositionIndex(CodesMessagePositionIndex):
    # This does not belong here, should be in the C library
    def _get_message_positions(self, path):
        fd = os.open(path, os.O_RDONLY)
        try:

            def get(count):
                buf = os.read(fd, count)
                assert len(buf) == count
                return int.from_bytes(
                    buf,
                    byteorder="big",
                    signed=False,
                )

            offset = 0
            while True:
                code = os.read(fd, 4)
                if len(code) < 4:
                    break

                if code != b"BUFR":
                    offset = os.lseek(fd, offset + 1, os.SEEK_SET)
                    continue

                length = get(3)
                edition = get(1)

                if edition in [3, 4]:
                    yield offset, length
                    offset = os.lseek(fd, offset + length, os.SEEK_SET)

        finally:
            os.close(fd)


class BUFRCodesHandle(CodesHandle):
    PRODUCT_ID = eccodes.CODES_PRODUCT_BUFR

    def __init__(self, handle, path, offset):
        super().__init__(handle, path, offset)
        self._unpacked = False

    def unpack(self):
        """Decode data section"""
        if not self._unpacked:
            eccodes.codes_set(self._handle, "unpack", 1)
            self._unpacked = True

    def pack(self):
        """Encode data section"""
        if self._unpacked:
            eccodes.codes_set(self._handle, "pack", 1)
            self._unpacked = False

    def json_dump(self, path):
        self.unpack()
        with open(path, "w") as f:
            eccodes.codes_dump(self._handle, f, "json")
        self.pack()

    def __iter__(self):
        class _KeyIterator:
            def __init__(self, handle):
                self._iterator = eccodes.codes_bufr_keys_iterator_new(handle)

            def __del__(self):
                try:
                    eccodes.codes_bufr_keys_iterator_delete(self._iterator)
                except Exception:
                    pass

            def __iter__(self):
                return self

            def __next__(self):
                while True:
                    if not eccodes.codes_bufr_keys_iterator_next(self._iterator):
                        raise StopIteration

                    return eccodes.codes_bufr_keys_iterator_get_name(self._iterator)

        return _KeyIterator(self._handle)

    def keys(self, namespace=None):
        """Iterate over all the available keys"""
        return self.__iter__()

    def as_namespace(self, namespace=None):
        return {k: self.get(k) for k in self.keys(namespace=namespace)}


class BUFRCodesReader(CodesReader):
    PRODUCT_ID = eccodes.CODES_PRODUCT_BUFR
    HANDLE_TYPE = BUFRCodesHandle


class BUFRMessage(Base):
    r"""Represents a BUFR message in a BUFR file.

    Parameters
    ----------
    path: str
        Path to the BUFR file
    offset: number
        File offset of the message (in bytes)
    length: number
        Size of the message (in bytes)
    """

    def __init__(self, path, offset, length):
        self.path = path
        self._offset = offset
        self._length = length
        self._handle = None

    @property
    def handle(self):
        r""":class:`CodesHandle`: Gets an object providing access to the low level BUFR message structure."""
        if self._handle is None:
            assert self._offset is not None
            self._handle = BUFRCodesReader.from_cache(self.path).at_offset(self._offset)
        return self._handle

    def __repr__(self):
        return "BUFRMessage(type=%s,subType=%s,subsets=%s,%s,%s)" % (
            self.handle.get("dataCategory", default=None),
            self.handle.get("dataSubCategory", default=None),
            self.handle.get("numberOfSubsets", default=None),
            self.handle.get("typicalDate", default=None),
            self.handle.get("typicalTime", default=None),
        )

    def _header(self, key):
        return self.handle.get(key, default=None)

    def subset_count(self):
        """Returns the number of subsets in the given BUFR message."""
        return self._header("numberOfSubsets")

    def is_compressed(self):
        """Checks if the BUFR message contains compressed subsets."""
        return self.subset_count() > 1 and self._header("compressedData") == 1

    def is_uncompressed(self):
        """Checks if the BUFR message contains uncompressed subsets."""
        return self.subset_count() > 1 and self._header("compressedData") == 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._handle = None

    def __setitem__(self, key, value):
        """Sets value associated with ``key``"""
        if isinstance(value, list):
            return eccodes.codes_set_array(self.handle._handle, key, value)
        else:
            return eccodes.codes_set(self.handle._handle, key, value)

    def __getitem__(self, key):
        """Returns the value of the ``key``."""
        return self.handle.get(key)

    def __iter__(self):
        """Returns an iterator for the keys the message contains."""
        return self.handle.__iter__()

    def unpack(self):
        """Decodes the data section of the message. When a message is unpacked all
        the keys in the data section become available via :obj:`metadata`.

        See Also
        --------
        :obj:`unpack`
        """
        return self.handle.unpack()

    def pack(self):
        """Encodes the data section of the message. Having called ``pack`` the
        contents of only the header keys become available via :obj:`metadata`. To access
        the data section you need to use :obj:`unpack` again.

        See Also
        --------
        :obj:`unpack`
        """
        return self.handle.unpack()

    def metadata(self, *keys, astype=None, **kwargs):
        r"""Returns metadata values from the BUFR message. When the message in packed
        (default state) only the header keys are available. To access the data section keys
        you need to call :obj:`unpack`.

        Parameters
        ----------
        *keys: tuple
            Positional arguments specifying metadata keys. Only ecCodes BUFR keys can be used
            here. It can contain a single str or a list or tuple. Can be empty, in this case
            all the keys will
            be used.
        astype: type name, :obj:`list` or :obj:`tuple`
            Return types for ``keys``. A single value is accepted and applied to all the ``keys``.
            Otherwise, must have same the number of elements as ``keys``. Only used when
            ``keys`` is not empty.
        **kwargs: tuple, optional
            Other keyword arguments:

            * default: value, optional
                Specifies the same default value for all the ``keys`` specified. When ``default`` is
                **not present** and a key is not found or its value is a missing value
                :obj:`metadata` will raise KeyError.

        Returns
        -------
        single value, :obj:`list`, :obj:`tuple` or :obj:`dict`
            - when ``keys`` is not empty:
                - single value when ``keys`` is a str
                - otherwise the same type as that of ``keys`` (:obj:`list` or :obj:`tuple`)
            - when ``keys`` is empty:
                - returns a :obj:`dict` with one item per key

        Raises
        ------
        KeyError
            If no ``default`` is set and a key is not found in the message or it has a missing value.

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/temp_10.bufr")
        >>> ds[0].metadata("edition")
        3
        >>> ds[0].metadata("dataCategory", "dataSubCategory")
        (2, 101)

        """
        key, namespace, astype, key_arg_type = metadata_argument(
            *keys, namespace=None, astype=astype
        )

        assert isinstance(key, list)
        assert isinstance(namespace, (list, tuple))

        if key:
            assert isinstance(astype, (list, tuple))
            r = [self.handle.get(k, ktype=kt, **kwargs) for k, kt in zip(key, astype)]

            if key_arg_type == str:
                return r[0]
            elif key_arg_type == tuple:
                return tuple(r)
            else:
                return r
        else:
            return self.handle.as_namespace()

    def is_coord(self, key):
        """Check if the specified key is a BUFR coordinate descriptor

        Parameters
        ----------
        key: str
            Key name (can contain ecCodes rank)

        Returns
        -------
        bool
            True if the specified ``key`` is a BUFR coordinate descriptor
        """
        try:
            return bufr_code_is_coord(self.d[key + "->code"])
        except Exception:
            return False

    def dump(self, subset=1):
        r"""Generates a dump with the message content represented as a tree view in a Jupyter notebook.

        Parameters
        ----------
        subset: int
            Subset to dump. Please note that susbset indexing starts at 1. Use None to dump all the
            subsets in the message.

        Returns
        -------
        HTML
            Dump contents represented as a tree view in a Jupyter notebook.

        Examples
        --------
        :ref:`/examples/bufr_temp.ipynb`

        """
        from earthkit.data.core.temporary import temp_file

        with temp_file() as filename:
            self.handle.json_dump(filename)
            with open(filename, "r") as f:
                import json
                import warnings

                try:
                    d = json.loads(f.read())
                    return make_bufr_html_tree(
                        d,
                        self.__repr__(),
                        subset,
                        self.is_compressed(),
                        self.is_uncompressed(),
                    )
                except Exception as e:
                    warnings.warn("Failed to parse bufr_dump", e)
                    return None

    def write(self, f):
        r"""Writes the message to a file object.

        Parameters
        ----------
        f: file object
            The target file object.
        """
        self.handle.write_to(f)

    def message(self):
        r"""Returns a buffer containing the encoded message.

        Returns
        -------
        bytes
        """
        return self.handle.get_buffer()


class BUFRListMixIn(PandasMixIn):
    def ls(self, *args, **kwargs):
        r"""Generates a list like summary of the BUFR message list using a set of metadata keys.

        Parameters
        ----------
        n: int, None
            The number of :obj:`BUFRMEssage <data.readers.bufr.bufr.BUFRMessage>`\ s to be
            listed. ``None`` means all the messages, ``n > 0`` means messages from the front, while
            ``n < 0`` means messages from the back of the list.
        keys: list of str, dict, None
            Metadata keys. To specify a column title for each key in the output use a dict. If
            ``keys`` is None the following dict will be used to define the titles and the keys::

                {
                    "edition": "edition",
                    "type": "dataCategory",
                    "subtype": "dataSubCategory",
                    "c": "bufrHeaderCentre",
                    "mv": "masterTablesVersionNumber",
                    "lv": "localTablesVersionNumber",
                    "subsets": "numberOfSubsets",
                    "compr": "compressedData",
                    "typicalDate": "typicalDate",
                    "typicalTime": "typicalTime",
                    "ident": "ident",
                    "lat": "localLatitude",
                    "lon": "localLongitude",
                }

        extra_keys: list of str, dict, None
            List of additional keys to ``keys``. To specify a column title for each key in the output
            use a dict.

        Returns
        -------
        Pandas DataFrame
            DataFrame with one row per :obj:`BUFRMEssage <data.readers.bufr.bufr.BUFRMessage>`.

        Examples
        --------
        :ref:`/examples/bufr_temp.ipynb`

        """
        from earthkit.data.utils.summary import ls

        def _proc(keys, n):
            count_start = 0
            if n is None:
                count_end = len(self)
            elif n > 0:
                count_end = n
            else:
                num = len(self)
                count_start = max(0, num + n)
                count_end = num

            for count, msg in enumerate(self):
                if count_start <= count < count_end:
                    yield ({k: msg._header(k1) for k, k1 in keys.items()})
                elif count >= count_end:
                    break

        return ls(_proc, BUFR_LS_KEYS, *args, **kwargs)

    def head(self, n=5, **kwargs):
        r"""Generates a list like summary of the first ``n``
        :obj:`BUFRMEssage <data.readers.bufr.bufr.BUFRMessage>`\ s using a set of metadata keys.
        Same as calling :obj:`ls` with ``n``.

        Parameters
        ----------
        n: int, None
            The number of messages (``n`` > 0) to be printed from the front.
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`ls`.

        Returns
        -------
        Pandas DataFrame
            See  :obj:`ls`.


        The following calls are equivalent:

            .. code-block:: python

                ds.head()
                ds.head(5)
                ds.head(n=5)
                ds.ls(5)
                ds.ls(n=5)

        """
        if n <= 0:
            raise ValueError("head: n must be > 0")
        return self.ls(n=n, **kwargs)

    def tail(self, n=5, **kwargs):
        r"""Generates a list like summary of the last ``n``
        :obj:`BUFRMEssage <data.readers.bufr.bufr.BUFRMessage>`\ s using a set of metadata keys.
        Same as calling :obj:`ls` with ``-n``.

        Parameters
        ----------
        n: int, None
            The number of messages (``n`` > 0)  to be printed from the back.
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`ls`.

        Returns
        -------
        Pandas DataFrame
            See  :obj:`ls`.


        The following calls are equivalent:

            .. code-block:: python

                ds.tail()
                ds.tail(5)
                ds.tail(n=5)
                ds.ls(-5)
                ds.ls(n=-5)

        """
        if n <= 0:
            raise ValueError("n must be > 0")
        return self.ls(n=-n, **kwargs)

    def metadata(self, *args, **kwargs):
        r"""Returns the metadata values for each message.

        Parameters
        ----------
        *args: tuple
            Positional arguments defining the metadata keys. Passed to
            :obj:`BUFRMessage.metadata() <data.readers.bufr.bufr.BUFRMessage.metadata>`
        **kwargs: dict, optional
            Keyword arguments passed to
            :obj:`BUFRMessage.metadata() <data.readers.bufr.bufr.BUFRMessage.metadata>`

        Returns
        -------
        list
            List with one item per :obj:`BUFRMessage <data.readers.bufr.bufr.BUFRMessage.metadata>`

        """
        result = []
        for s in self:
            result.append(s.metadata(*args, **kwargs))
        return result


class BUFRList(BUFRListMixIn, Index):
    r"""Represents a list of
    :obj:`BUFRMessage <data.readers.bufr.bufr.BUFRMessage>`\ s.
    """

    def __init__(self, *args, **kwargs):
        Index.__init__(self, *args, **kwargs)

    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return MaskBUFRList(*args, **kwargs)

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(_, BUFRList) for _ in sources)
        return MultiBUFRList(sources)

    def sel(self, *args, **kwargs):
        """Uses header metadata values to select only certain messages from a BUFRList object.

        Parameters
        ----------
        *args: tuple
            Positional arguments specifying the filter condition as dict.
            (See below for details).
        **kwargs: dict, optional
            Other keyword arguments specifying the filter conditions.
            (See below for details).

        Returns
        -------
        object
            Returns a new object with the filtered elements. It contains a view to the data in the
            original object, so no data is copied.


        Filter conditions are specified by a set of **metadata** keys either by a dictionary (in
        ``*args``) or a set of ``**kwargs``. Both single or multiple keys are allowed to use and each
        can specify the following type of filter values:

        - single value::

            ds.sel(dataCategory="2")

        - list of values::

            ds.sel(dataCategory=[1, 2])

        - **slice** of values (defines a **closed interval**, so treated as inclusive of both the start
        and stop values, unlike normal Python indexing)::

            # filter dataCategory between 1 and 4 inclusively
            ds.sel(dataCategory=slice(1,4))
        """
        kwargs.pop("remapping", None)
        return super().sel(*args, remapping=None, **kwargs)

    def isel(self, *args, **kwargs):
        raise NotImplementedError()

    def order_by(self, *args, **kwargs):
        """Changes the order of the messages in a BUFRList object.

        Parameters
        ----------
        *args: tuple
            Positional arguments specifying the metadata keys to perform the ordering on.
            (See below for details)
        **kwargs: dict, optional
            Other keyword arguments specifying the metadata keys to perform the ordering on.
            (See below for details)

        Returns
        -------
        object
            Returns a new object with reordered messages. It contains a view to the data in the
            original object, so no data is copied.

        """
        kwargs.pop("remapping", None)
        return super().order_by(*args, remapping=None, **kwargs)


class MaskBUFRList(BUFRList, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)


class MultiBUFRList(BUFRList, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)


class BUFRInFiles(BUFRList):
    # Remote BUFRLists (with urls) are also here,
    # as the actual fieldlist is accessed on a file in cache.
    # This class changes the interface (_getitem__ and __len__)
    # into the interface (part and number_of_parts).
    def __getitem__(self, n):
        if isinstance(n, int):
            part = self.part(n if n >= 0 else len(self) + n)
            return BUFRMessage(part.path, part.offset, part.length)
        else:
            return super().__getitem__(n)

    def __len__(self):
        return self.number_of_parts()

    @abstractmethod
    def part(self, n):
        self._not_implemented()

    @abstractmethod
    def number_of_parts(self):
        self._not_implemented()


class BUFRInOneFile(BUFRInFiles):
    def __init__(self, path):
        self.path = path
        self.__positions = None

    @property
    def _positions(self):
        if self.__positions is None:
            self.__positions = BufrCodesMessagePositionIndex(self.path)
        return self.__positions

    def part(self, n):
        return Part(self.path, self._positions.offsets[n], self._positions.lengths[n])

    def number_of_parts(self):
        return len(self._positions)


class BUFRReader(BUFRInOneFile, Reader):
    """Represents a BUFR file"""

    appendable = True  # BUFR messages can be added to the same file

    def __init__(self, source, path):
        Reader.__init__(self, source, path)
        BUFRInOneFile.__init__(self, path)

    def __repr__(self):
        return "BUFRReader(%s)" % (self.path,)

    @classmethod
    def merge(cls, readers):
        assert all(isinstance(s, BUFRReader) for s in readers), readers
        assert len(readers) > 1

        return MultiBUFRList(readers)

    def mutate_source(self):
        # A BUFRReader is a source itself
        return self



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

LOG = logging.getLogger(__name__)

COLUMNS = ("latitude", "longitude", "data_datetime")


class PandasMixIn:
    def to_pandas(self, columns=COLUMNS, filters=None, **kwargs):
        """Extracts BUFR data into a pandas DataFrame using :xref:`pdbufr`.

        Parameters
        ----------
        columns: str, sequence[str]
            List of ecCodes BUFR keys to extract for each BUFR message/subset.
            See: :xref:`read_bufr` for details.
        filters: dict
            Defines the conditions when to extract the specified ``columns``. See:
            :xref:`read_bufr` for details.
        **kwargs: dict, optional
            Other keyword arguments passed to :xref:`read_bufr`.

        Returns
        -------
        Pandas DataFrame

        Examples
        --------
        - :ref:`/examples/bufr_temp.ipynb`
        - :ref:`/examples/bufr_synop.ipynb`

        """
        import pdbufr

        filters = {} if filters is None else filters

        return pdbufr.read_bufr(self, columns=columns, filters=filters, **kwargs)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file, earthkit_test_data_file


def check_array(v, shape=None, first=None, last=None, meanv=None, eps=1e-3):
    assert v.shape == shape
    assert np.isclose(v[0], first, eps)
    assert np.isclose(v[-1], last, eps)
    assert np.isclose(v.mean(), meanv, eps)


def test_grib_values_1():
    f = from_source("file", earthkit_test_data_file("test_single.grib"))

    eps = 1e-5

    # whole file
    v = f.values
    assert isinstance(v, np.ndarray)
    assert v.shape == (1, 84)
    v = v[0].flatten()
    check_array(
        v,
        (84,),
        first=260.43560791015625,
        last=227.18560791015625,
        meanv=274.36566743396577,
        eps=eps,
    )

    # field
    v1 = f[0].values
    assert isinstance(v1, np.ndarray)
    assert v1.shape == (84,)
    assert np.allclose(v, v1, eps)


def test_grib_values_18():
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))

    eps = 1e-5

    # whole file
    v = f.values
    assert isinstance(v, np.ndarray)
    assert v.shape == (18, 84)
    vf = v[0].flatten()
    check_array(
        vf,
        (84,),
        first=272.5642,
        last=240.56417846679688,
        meanv=279.70703560965404,
        eps=eps,
    )

    vf = v[15].flatten()
    check_array(
        vf,
        (84,),
        first=226.6531524658203,
        last=206.6531524658203,
        meanv=227.84362865629652,
        eps=eps,
    )


def test_grib_to_numpy_1():
    f = from_source("file", earthkit_test_data_file("test_single.grib"))

    eps = 1e-5
    v = f.to_numpy()
    assert isinstance(v, np.ndarray)
    v = v[0].flatten()
    check_array(
        v,
        (84,),
        first=260.43560791015625,
        last=227.18560791015625,
        meanv=274.36566743396577,
        eps=eps,
    )


@pytest.mark.parametrize(
    "first,options, expected_shape",
    [
        (False, {}, (1, 7, 12)),
        (False, {"flatten": True}, (1, 84)),
        (False, {"flatten": False}, (1, 7, 12)),
        (True, {}, (7, 12)),
        (True, {"flatten": True}, (84,)),
        (True, {"flatten": False}, (7, 12)),
    ],
)
def test_grib_to_numpy_1_shape(first, options, expected_shape):
    f = from_source("file", earthkit_test_data_file("test_single.grib"))

    v_ref = f[0].to_numpy().flatten()
    eps = 1e-5

    data = f[0] if first else f
    v1 = data.to_numpy(**options)
    assert isinstance(v1, np.ndarray)
    assert v1.shape == expected_shape
    v1 = v1.flatten()
    assert np.allclose(v_ref, v1, eps)


def test_grib_to_numpy_18():
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))

    eps = 1e-5

    # whole file
    v = f.to_numpy(flatten=True)
    assert isinstance(v, np.ndarray)
    assert v.shape == (18, 84)
    vf0 = v[0].flatten()
    check_array(
        vf0,
        (84,),
        first=272.5642,
        last=240.56417846679688,
        meanv=279.70703560965404,
        eps=eps,
    )

    vf15 = v[15].flatten()
    check_array(
        vf15,
        (84,),
        first=226.6531524658203,
        last=206.6531524658203,
        meanv=227.84362865629652,
        eps=eps,
    )


@pytest.mark.parametrize(
    "options, expected_shape",
    [
        (
            {},
            (
                18,
                7,
                12,
            ),
        ),
        (
            {"flatten": True},
            (
                18,
                84,
            ),
        ),
        ({"flatten": False}, (18, 7, 12)),
    ],
)
def test_grib_to_numpy_18_shape(options, expected_shape):
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))

    eps = 1e-5

    # whole file
    v = f.to_numpy()
    assert isinstance(v, np.ndarray)
    assert v.shape == (18, 7, 12)
    vf0 = f[0].to_numpy().flatten()
    assert vf0.shape == (84,)
    vf15 = f[15].to_numpy().flatten()
    assert vf15.shape == (84,)

    v1 = f.to_numpy(**options)
    assert isinstance(v1, np.ndarray)
    assert v1.shape == expected_shape
    vr = v1[0].flatten()
    assert np.allclose(vf0, vr, eps)
    vr = v1[15].flatten()
    assert np.allclose(vf15, vr, eps)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_grib_to_numpy_1_dtype(dtype):
    f = from_source("file", earthkit_test_data_file("test_single.grib"))

    v = f[0].to_numpy(dtype=dtype)
    assert v.dtype == dtype

    v = f.to_numpy(dtype=dtype)
    assert v.dtype == dtype


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_grib_to_numpy_18_dtype(dtype):
    f = from_source("file", earthkit_examples_file("tuv_pl.grib"))

    v = f[0].to_numpy(dtype=dtype)
    assert v.dtype == dtype

    v = f.to_numpy(dtype=dtype)
    assert v.dtype == dtype


@pytest.mark.parametrize(
    "kwarg,expected_shape",
    [({}, (11, 19)), ({"flatten": True}, (209,)), ({"flatten": False}, (11, 19))],
)
def test_grib_field_data(kwarg, expected_shape):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    latlon = ds[0].to_latlon(**kwarg)
    v = ds[0].to_numpy(**kwarg)

    d = ds[0].data(**kwarg)
    assert isinstance(d, np.ndarray)
    assert len(d) == 3
    assert d[0].shape == expected_shape
    assert np.allclose(d[0], latlon["lat"])
    assert np.allclose(d[1], latlon["lon"])
    assert np.allclose(d[2], v)

    d = ds[0].data(keys="lat", **kwarg)
    assert d.shape == expected_shape
    assert np.allclose(d, latlon["lat"])

    d = ds[0].data(keys="lon", **kwarg)
    assert d.shape == expected_shape
    assert np.allclose(d, latlon["lon"])

    d = ds[0].data(keys="value", **kwarg)
    assert d.shape == expected_shape
    assert np.allclose(d, v)

    d = ds[0].data(keys=("value", "lon"), **kwarg)
    assert isinstance(d, np.ndarray)
    assert len(d) == 2
    assert np.allclose(d[0], v)
    assert np.allclose(d[1], latlon["lon"])


@pytest.mark.parametrize(
    "kwarg,expected_shape",
    [
        ({}, (11, 19)),
        (
            {"flatten": True},
            (209,),
        ),
        ({"flatten": False}, (11, 19)),
    ],
)
def test_grib_fieldlist_data(kwarg, expected_shape):
    ds = from_source("file", earthkit_examples_file("test.grib"))

    latlon = ds.to_latlon(**kwarg)
    v = ds.to_numpy(**kwarg)

    d = ds.data(**kwarg)
    assert isinstance(d, np.ndarray)
    assert d.shape == tuple([4, *expected_shape])
    assert np.allclose(d[0], latlon["lat"])
    assert np.allclose(d[1], latlon["lon"])
    assert np.allclose(d[2], v[0])
    assert np.allclose(d[3], v[1])

    d = ds.data(keys="lat", **kwarg)
    assert d.shape == tuple([1, *expected_shape])
    assert np.allclose(d[0], latlon["lat"])

    d = ds.data(keys="lon", **kwarg)
    assert d.shape == tuple([1, *expected_shape])
    assert np.allclose(d[0], latlon["lon"])

    d = ds.data(keys="value", **kwarg)
    assert d.shape == tuple([2, *expected_shape])
    assert np.allclose(d, v)

    d = ds.data(keys=("value", "lon"), **kwarg)
    assert isinstance(d, np.ndarray)
    assert d.shape == tuple([3, *expected_shape])
    assert np.allclose(d[0], v[0])
    assert np.allclose(d[1], v[1])
    assert np.allclose(d[2], latlon["lon"])


def test_grib_values_with_missing():
    f = from_source("file", earthkit_test_data_file("test_single_with_missing.grib"))

    v = f[0].values
    assert isinstance(v, np.ndarray)
    assert v.shape == (84,)
    eps = 0.001
    assert np.count_nonzero(np.isnan(v)) == 38
    mask = np.array([12, 14, 15, 24, 25, 26] + list(range(28, 60)))
    assert np.isclose(v[0], 260.4356, eps)
    assert np.isclose(v[11], 260.4356, eps)
    assert np.isclose(v[-1], 227.1856, eps)
    m = v[mask]
    assert len(m) == 38
    assert np.count_nonzero(np.isnan(m)) == 38


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_file


# @pytest.mark.skipif(("GITHUB_WORKFLOW" in os.environ) or True, reason="Not yet ready")
def test_grib_order_by_single_message():
    s = from_source("file", earthkit_file("tests/data/test_single.grib"))

    r = s.order_by("shortName")
    assert len(r) == 1
    assert r[0].metadata("shortName") == "2t"

    r = s.order_by(["shortName"])
    assert len(r) == 1
    assert r[0].metadata("shortName") == "2t"

    r = s.order_by(["shortName", "level"])
    assert len(r) == 1
    assert r[0].metadata("shortName") == "2t"


class _CustomOrder:
    def __call__(self, x, y):
        r = dict(t=2, u=0, v=1)
        a = r[x]
        b = r[y]
        if a == b:
            return 0
        if a > b:
            return 1
        if a < b:
            return -1


@pytest.mark.parametrize(
    "params,expected_meta",
    [
        ("shortName", dict(shortName=["t", "t", "u", "u", "v", "v"])),
        (["shortName"], dict(shortName=["t", "t", "u", "u", "v", "v"])),
        (
            ["shortName", "level"],
            dict(
                shortName=["t", "t", "u", "u", "v", "v"],
                level=[850, 1000, 850, 1000, 850, 1000],
            ),
        ),
        (dict(shortName="ascending"), dict(shortName=["t", "t", "u", "u", "v", "v"])),
        (dict(shortName="descending"), dict(shortName=["v", "v", "u", "u", "t", "t"])),
        (
            dict(shortName="ascending", level="ascending"),
            dict(
                shortName=["t", "t", "u", "u", "v", "v"],
                level=[850, 1000, 850, 1000, 850, 1000],
            ),
        ),
        (
            dict(shortName="ascending", level="descending"),
            dict(
                shortName=["t", "t", "u", "u", "v", "v"],
                level=[1000, 850, 1000, 850, 1000, 850],
            ),
        ),
        (
            dict(shortName=_CustomOrder()),
            dict(shortName=["u", "u", "v", "v", "t", "t"]),
        ),
        (
            dict(shortName=["u", "v", "t"]),
            dict(shortName=["u", "u", "v", "v", "t", "t"]),
        ),
        (
            dict(shortName=["u", "v", "t"], level=[1000, 850]),
            dict(
                shortName=["u", "u", "v", "v", "t", "t"],
                level=[1000, 850, 1000, 850, 1000, 850],
            ),
        ),
    ],
)
def test_grib_order_by_single_file_(
    params,
    expected_meta,
):
    f = from_source("file", earthkit_file("docs/examples/test6.grib"))

    g = f.order_by(params)
    assert len(g) == len(f)

    for k, v in expected_meta.items():
        assert g.metadata(k) == v


@pytest.mark.parametrize(
    "params,expected_meta",
    [
        (
            "shortName",
            dict(shortName=["t", "t", "t", "t", "u", "u", "v", "v", "z", "z"]),
        ),
        (
            ["shortName"],
            dict(shortName=["t", "t", "t", "t", "u", "u", "v", "v", "z", "z"]),
        ),
        (
            ["shortName", "level"],
            dict(
                shortName=["t", "t", "t", "t", "u", "u", "v", "v", "z", "z"],
                level=[500, 850, 850, 1000, 850, 1000, 850, 1000, 500, 850],
            ),
        ),
        (
            dict(shortName="ascending"),
            dict(shortName=["t", "t", "t", "t", "u", "u", "v", "v", "z", "z"]),
        ),
        (
            dict(shortName="descending"),
            dict(shortName=["z", "z", "v", "v", "u", "u", "t", "t", "t", "t"]),
        ),
    ],
)
def test_grib_order_by_multi_file(params, expected_meta):
    f1 = from_source("file", earthkit_file("docs/examples/test4.grib"))
    f2 = from_source("file", earthkit_file("docs/examples/test6.grib"))
    f = from_source("multi", [f1, f2])

    g = f.order_by(params)
    assert len(g) == len(f)

    for k, v in expected_meta.items():
        assert g.metadata(k) == v


def test_grib_order_by_with_sel():
    f = from_source("file", earthkit_file("docs/examples/tuv_pl.grib"))

    g = f.sel(level=500)
    assert len(g) == 3
    r = g.order_by("shortName")
    assert len(r) == len(g)
    assert r.metadata("shortName") == ["t", "u", "v"]

    g = f.sel(level=500)
    assert len(g) == 3
    r = g.order_by({"shortName": "descending"})
    assert len(r) == len(g)
    assert r.metadata("shortName") == ["v", "u", "t"]


def test_grib_order_by_valid_datetime():
    f = from_source("file", earthkit_file("tests/data/t_time_series.grib"))

    g = f.order_by(valid_datetime="descending")
    assert len(g) == 10

    ref = [
        datetime.datetime(2020, 12, 23, 12, 0),
        datetime.datetime(2020, 12, 23, 12, 0),
        datetime.datetime(2020, 12, 21, 21, 0),
        datetime.datetime(2020, 12, 21, 21, 0),
        datetime.datetime(2020, 12, 21, 18, 0),
        datetime.datetime(2020, 12, 21, 18, 0),
        datetime.datetime(2020, 12, 21, 15, 0),
        datetime.datetime(2020, 12, 21, 15, 0),
        datetime.datetime(2020, 12, 21, 12, 0),
        datetime.datetime(2020, 12, 21, 12, 0),
    ]

    assert g.metadata("valid_datetime") == ref



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import ALL
from earthkit.data.arguments.earthkit_types import EnumListType, EnumType
from earthkit.data.decorators import normalize


def name_no_default(name):
    return name


def name_default_is_none(name=None):
    return name


def name_default_is_1(name=1):
    return name


def name_default_is_str_1(name="1"):
    return name


def name_default_is_str_a(name="a"):
    return name


def name_default_is_all(name=ALL):
    return name


def a_b_no_default(a, b):
    return a, b


def test_enum_2_normalizers():
    g = a_b_no_default
    g = normalize("a", [1, 2])(g)
    g = normalize("b", [3, 4])(g)
    assert g(a=1, b=4) == (1, 4)
    with pytest.raises(TypeError):
        g(a=1)


def test_enum_decorator():
    g = normalize("name", ("a", "b", "c"))(name_default_is_str_a)
    assert g("a") == "a"
    assert g("b") == "b"
    assert g() == "a"
    with pytest.raises(ValueError):
        g("z")
    assert g(["a", "b"]) == ["a", "b"]
    assert g(("a", "b")) == ("a", "b")
    assert g(["a"]) == ["a"]


def test_enum_multiple():
    g = normalize("name", ["a", "b", "c"], multiple=False)(name_default_is_str_a)
    with pytest.raises(TypeError):
        g(["a", "b"])


def test_enum_multiple_2():
    g = normalize("name", ["a", "b", "c"], multiple=True)(name_default_is_str_a)
    assert g(
        (
            "a",
            "b",
        )
    ) == ["a", "b"]


def test_enum_multiple_ALL_1():
    g = normalize("name", ["a", "b", "c"], multiple=False)(name_default_is_str_a)
    with pytest.raises(ValueError):
        g(ALL)


def test_enum_multiple_ALL_2():
    g = normalize("name", ["a", "b", "c"], multiple=True)(name_default_is_str_a)
    assert g(ALL) == ["a", "b", "c"]


def test_enum_float_1():
    g = normalize("name", [1, 0.5, 3], type=float, format="%03f")(name_default_is_1)
    assert g(1) == "1.000000"


@pytest.mark.skip("Not implemented. Need to discuss what it would mean.")
def test_enum_float_2():
    g = normalize("name", type=EnumListType([1, 0.5, 3]), format="%03f")(
        name_no_default
    )
    assert g(1) == ["1.000000"]

    g = normalize("name", type=EnumListType([1, 0.5, 3]), format="%03f", multiple=True)(
        name_no_default
    )
    assert g(1) == ["1.000000"]

    g = normalize(
        "name", type=EnumListType([1, 0.5, 3]), format="%03f", multiple=False
    )(name_no_default)
    with pytest.raises(ValueError, match="Cannot .*"):
        assert g(1) == ["1.000000"]


@pytest.mark.skip("Not implemented. Need to discuss what it would mean.")
def test_enum_float_3():
    g = normalize("name", type=EnumType([1, 0.5, 3]), format="%03f")(name_no_default)
    assert g(1) == "1.000000"

    g = normalize("name", type=EnumType([1, 0.5, 3]), format="%03f", multiple=False)(
        name_no_default
    )
    assert g(1) == "1.000000"

    g = normalize("name", type=EnumType([1, 0.5, 3]), format="%03f", multiple=True)(
        name_no_default
    )
    with pytest.raises(ValueError, match="Cannot .*"):
        assert g(1) == ["1.000000"]


def test_enum_int_1():
    g = normalize("name", [1, 0.5, 3], type=int, multiple=True)(name_default_is_1)
    assert g(1) == [1]
    assert g(1.0) == [1]
    # assert g("1.0") == [1]
    assert g("1") == [1]
    assert g() == [1]


def test_enum_int_2():
    g = normalize("name", (1, 2, 3))(name_no_default)
    assert g("1") == 1
    assert g(1.0) == 1
    # assert g("1.0") == 1


def test_enum_int_3():
    g = normalize("name", [1, 2, 3], multiple=True)(name_no_default)
    assert g("1") == [1]
    assert g(1.0) == [1]
    # assert g("1.0") == [1]


def test_enum_float():
    g = normalize("name", (1.0, 0.5, 3.0))(name_no_default)
    assert g(1) == 1.0
    assert g("1") == 1.0


def test_enum_list_case_sensitive():
    g = normalize("name", ["A", "b", "c"], multiple=True)(name_no_default)
    # TODO:    assert g(ALL) == ["A", "b", "c"]
    assert g("a") == ["A"]
    assert g("A") == ["A"]
    assert g(["a", "B"]) == ["A", "b"]


def test_enum_list_alias_1():
    enum_list_alias_1 = normalize(
        "name",
        ["a", "b", "c"],
        multiple=True,
        aliases={
            "ab": ["a", "b"],
            "z": "a",
            "i": ["a", "b"],
            "j": "ab",
            "bad": ["a", "ab"],
        },
    )(name_no_default)
    assert enum_list_alias_1("a") == ["a"]
    assert enum_list_alias_1("b") == ["b"]
    assert enum_list_alias_1("ab") == ["a", "b"]
    assert enum_list_alias_1("z") == ["a"]
    assert enum_list_alias_1(["z", "b"]) == ["a", "b"]
    assert enum_list_alias_1("i") == ["a", "b"]
    assert enum_list_alias_1("j") == ["a", "b"]
    # TODO: add more check for bad aliases in normalize
    # with pytest.raises(ValueError):
    #    enum_list_alias_1("bad")


def test_enum_list_alias_2():
    enum_list_alias_2 = normalize(
        "name",
        [1, 2, 3],
        multiple=True,
        aliases=lambda x: {"one": 1, "o": "one"}.get(x, x),
    )(name_no_default)
    assert enum_list_alias_2(1) == [1]
    assert enum_list_alias_2("one") == [1]
    assert enum_list_alias_2(["one"]) == [1]
    assert enum_list_alias_2(["o"]) == [1]


def test_enum_alias():
    enum_alias = normalize(
        "name",
        ["a", "b", "c"],
        multiple=True,
        aliases={"x": "y", "y": "z", "z": "a"},
    )(name_no_default)
    assert enum_alias("a") == ["a"]
    assert enum_alias("b") == ["b"]
    assert enum_alias("x") == ["a"]
    assert enum_alias("y") == ["a"]
    assert enum_alias("z") == ["a"]


def test_enum_none_1():
    @normalize(
        "name",
        ["a", "b", "c"],
        multiple=True,
        aliases={"x": "y", "y": "z", "z": "a"},
    )
    def enum_none(name):
        return name

    assert enum_none(None) is None


def test_enum_none_2():
    @normalize("name", ["a", "b", "c"], multiple=True, aliases={None: "b"})
    def enum_none_2(name):
        return name

    assert enum_none_2(None) == ["b"]


def test_enum_none_3():
    @normalize("name", ["a", "b", "c"], aliases={None: "b"})
    def enum_none_3(name):
        return name

    assert enum_none_3(None) == "b"


def test_enum_none_4():
    @normalize("name", ["a", "b", "c"], aliases={None: "b"})
    def enum_none_4(name=None):
        return name

    assert enum_none_4() == "b"


def test_enum_alias_2():
    enum_alias = normalize(
        "name",
        ["a", "b", "c"],
        aliases={"x": "y", "y": "z", "z": "a", "w": "wrong-value"},
    )(name_no_default)
    assert enum_alias("a") == "a"
    assert enum_alias("b") == "b"
    assert enum_alias("x") == "a"
    assert enum_alias("y") == "a"
    assert enum_alias("z") == "a"
    with pytest.raises(ValueError, match=".*wrong-value.*"):
        enum_alias("w")


def test_enum_default_1():
    @normalize(
        "name",
        ["a", "b", "c"],
    )
    def enum_default_1(name="wrong-default"):
        return name

    enum_default_1("a")
    with pytest.raises(ValueError, match=".*wrong-default.*"):
        enum_default_1()


def test_enum_aliases_from_file():
    enum_aliases_from_file = normalize(
        "name",
        ["a", "b", "c"],
        aliases="aliases.json",
    )(name_no_default)
    assert enum_aliases_from_file("y") == "b"
    with pytest.raises(ValueError):
        enum_aliases_from_file("unknown")


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging

from earthkit.data import from_object, wrappers
from earthkit.data.wrappers import xarray as xr_wrapper

LOG = logging.getLogger(__name__)


def test_dataset_wrapper():
    import xarray as xr

    _wrapper = xr_wrapper.wrapper(xr.Dataset())
    assert isinstance(_wrapper, xr_wrapper.XArrayDatasetWrapper)
    _wrapper = wrappers.get_wrapper(xr.Dataset())
    assert isinstance(_wrapper, xr_wrapper.XArrayDatasetWrapper)
    _wrapper = from_object(xr.Dataset())
    assert isinstance(_wrapper, xr_wrapper.XArrayDatasetWrapper)


def test_dataarray_wrapper():
    import xarray as xr

    _wrapper = xr_wrapper.wrapper(xr.DataArray())
    assert isinstance(_wrapper, xr_wrapper.XArrayDataArrayWrapper)
    _wrapper = wrappers.get_wrapper(xr.DataArray())
    assert isinstance(_wrapper, xr_wrapper.XArrayDataArrayWrapper)
    _wrapper = from_object(xr.DataArray())
    assert isinstance(_wrapper, xr_wrapper.XArrayDataArrayWrapper)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import functools
import getpass
import logging
import os
import tempfile
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from typing import Callable

import yaml

from earthkit.data import __version__ as VERSION
from earthkit.data.utils.html import css
from earthkit.data.utils.humanize import (
    as_bytes,
    as_percent,
    as_seconds,
    interval_to_human,
    list_to_human,
)
from earthkit.data.utils.interval import Interval

LOG = logging.getLogger(__name__)

DOT_EARTHKIT_DATA = os.path.expanduser("~/.earthkit_data")
EARTHKIT_SETTINGS_DIR = DOT_EARTHKIT_DATA


class Validator(metaclass=ABCMeta):
    @abstractmethod
    def check(self, value):
        pass

    @abstractmethod
    def explain(self):
        pass

    @staticmethod
    def make(value):
        v = _validators.get(type(value), None)
        if v is not None:
            return v(value)
        else:
            raise TypeError(f"Cannot create Validator for type={type(value)}")


class ValueValidator(Validator):
    def __init__(self, value):
        self.value = value

    def check(self, value):
        return value == self.value

    def explain(self):
        return f"Valid when = {self.value}."


class IntervalValidator(Validator):
    def __init__(self, interval):
        self.interval = interval

    def check(self, value):
        return value in self.interval

    def explain(self):
        return f"Valid when {interval_to_human(self.interval)}."


class ListValidator(Validator):
    def __init__(self, values):
        self.values = values

    def check(self, value):
        return value in self.values

    def explain(self):
        return f"Valid values: {list_to_human(self.values)}."


_validators = {Interval: IntervalValidator, bool: ValueValidator, list: ListValidator}


class Setting:
    def __init__(
        self,
        default,
        description,
        getter=None,
        none_ok=False,
        kind=None,
        docs_default=None,
        validator=None,
    ):
        self.default = default
        self.description = description
        self.getter = getter
        self.none_ok = none_ok
        self.kind = kind if kind is not None else type(default)
        self.docs_default = docs_default if docs_default is not None else self.default
        self.validator = validator

    def kind(self):
        return type(self.default)

    def save(self, name, value, f):
        for n in self.description.split("\n"):
            print(f"# {n.strip()}", file=f)
        print(file=f)
        comment = yaml.dump({name: self.default}, default_flow_style=False)
        for n in comment.split("\n"):
            if n:
                print(f"# {n}", file=f)
        if value != self.default:
            print(file=f)
            yaml.dump({name: value}, f, default_flow_style=False)

    @property
    def docs_description(self):
        d = self.description
        t = ""
        if self.validator:
            t = self.validator.explain()

        return d.replace("{validator}", t)

    def validate(self, name, value):
        if self.validator is not None and not self.validator.check(value):
            raise ValueError(
                f"Settings {name} cannot be set to {value}. {self.validator.explain()}"
            )


_ = Setting


SETTINGS_AND_HELP = {
    "user-cache-directory": _(
        os.path.join(tempfile.gettempdir(), "earthkit-data-%s" % (getpass.getuser(),)),
        """Cache directory used when ``cache-policy`` is ``user``.
        See :doc:`/guide/caching` for more information.""",
        docs_default=os.path.join("TMP", "earthkit-data-%s" % (getpass.getuser(),)),
    ),
    "temporary-cache-directory-root": _(
        None,
        """Parent of the cache directory when ``cache-policy`` is ``temporary``.
        See :doc:`/guide/caching` for more information.""",
        getter="_as_str",
        none_ok=True,
    ),
    "number-of-download-threads": _(
        5,
        """Number of threads used to download data.""",
    ),
    "cache-policy": _(
        "user",
        """Caching policy. {validator} See :doc:`/guide/caching` for more information. """,
        validator=ListValidator(["off", "temporary", "user"]),
    ),
    "use-message-position-index-cache": _(
        False,
        "Stores message offset index for GRIB/BUFR files in the cache.",
    ),
    "maximum-cache-size": _(
        None,
        """Maximum disk space used by the earthkit-data cache (ex: 100G or 2T).""",
        getter="_as_bytes",
        none_ok=True,
    ),
    "maximum-cache-disk-usage": _(
        "95%",
        """Disk usage threshold after which earthkit-data expires older cached
        entries (% of the full disk capacity).
        See :doc:`/guide/caching` for more information.""",
        getter="_as_percent",
    ),
    "url-download-timeout": _(
        "30s",
        """Timeout when downloading from an url.""",
        getter="_as_seconds",
    ),
    "check-out-of-date-urls": _(
        True,
        "Perform a HTTP request to check if the remote version of a cache file has changed",
    ),
    "download-out-of-date-urls": _(
        False,
        "Re-download URLs when the remote version of a cached file as been changed",
    ),
    "use-standalone-mars-client-when-available": _(
        True,
        "Use the standalone mars client when available instead of using the web API.",
    ),
    "reader-type-check-bytes": _(
        64,
        "Number of bytes read from the beginning of a source to identify its type. {validator}",
        validator=IntervalValidator(Interval(8, 4096)),
    ),
}


NONE = object()
DEFAULTS = {}
for k, v in SETTINGS_AND_HELP.items():
    DEFAULTS[k] = v.default


@contextmanager
def new_settings(s):
    """Context manager to create new settings"""
    SETTINGS._stack.append(s)
    SETTINGS._notify()
    try:
        yield None
    finally:
        SETTINGS._stack.pop()
        SETTINGS._notify()


def forward(func):
    @functools.wraps(func)
    def wrapped(self, *args, **kwargs):
        if self._stack:
            return func(self._stack[-1], *args, **kwargs)
        return func(self, *args, **kwargs)

    return wrapped


def save_settings(path, settings):
    LOG.debug("Saving settings")
    with open(path, "w") as f:
        print("# This file is automatically generated", file=f)
        print(file=f)

        for k, v in sorted(settings.items()):
            h = SETTINGS_AND_HELP.get(k)
            if h:
                print(file=f)
                print("#", "-" * 76, file=f)
                h.save(k, v, f)

        print(file=f)
        print("#", "-" * 76, file=f)
        print("# Version of CliMetLab", file=f)
        print(file=f)
        yaml.dump({"version": VERSION}, f, default_flow_style=False)
        print(file=f)


class Settings:
    _auto_save_settings = True
    _notify_enabled = True

    def __init__(self, settings_yaml: str, defaults: dict, callbacks=[]):
        self._defaults = defaults
        self._settings = dict(**defaults)
        self._callbacks = [c for c in callbacks]
        self._settings_yaml = settings_yaml
        self._pytest = None
        self._stack = []

    @forward
    def get(self, name: str, default=NONE):
        """[summary]

        Parameters
        ----------
            name: str
                [description]
            default: [type]
                [description]. Defaults to NONE.

        Returns
        -------
            [type]: [description]
        """
        if name not in SETTINGS_AND_HELP:
            raise KeyError("No setting name '%s'" % (name,))

        settings_item = SETTINGS_AND_HELP[name]

        getter, none_ok = (
            settings_item.getter,
            settings_item.none_ok,
        )
        if getter is None:
            getter = lambda name, value, none_ok: value  # noqa: E731
        else:
            getter = getattr(self, getter)

        if default is NONE:
            return getter(name, self._settings[name], none_ok)

        return getter(name, self._settings.get(name, default), none_ok)

    @forward
    def set(self, *args, **kwargs):
        """[summary]

        Parameters
        ----------
            name: str
                [description]
            value: [type]
                [description]
        """
        if len(args) == 0:
            assert len(kwargs) > 0
            for k, v in kwargs.items():
                self._set(k.replace("_", "-"), v)
            self._changed()
        else:
            if len(args) == 1 and isinstance(args[0], dict):
                for k, v in args[0].items():
                    self._set(k, v)
            else:
                self._set(*args, **kwargs)
        self._changed()

    def _set(self, name: str, *args, **kwargs):
        """[summary]

        Parameters
        ----------
            name: str
                [description]
            value: [type]
                [description]
        """
        if name not in SETTINGS_AND_HELP:
            raise KeyError("No setting name '%s'" % (name,))

        settings_item = SETTINGS_AND_HELP[name]

        klass = settings_item.kind

        if klass in (bool, int, float, str):
            # TODO: Proper exceptions
            assert len(args) == 1
            assert len(kwargs) == 0
            value = args[0]
            value = klass(value)

        if klass is list:
            assert len(args) > 0
            assert len(kwargs) == 0
            value = list(args)
            if len(args) == 1 and isinstance(args[0], list):
                value = args[0]

        if klass is dict:
            assert len(args) <= 1
            if len(args) == 0:
                assert len(kwargs) > 0
                value = kwargs

            if len(args) == 1:
                assert len(kwargs) == 0
                value = args[0]

        getter, none_ok = (
            settings_item.getter,
            settings_item.none_ok,
        )
        if getter is not None:
            assert len(args) == 1
            assert len(kwargs) == 0
            value = args[0]
            # Check if value is properly formatted for getter
            getattr(self, getter)(name, value, none_ok)
        else:
            if not isinstance(value, klass):
                raise TypeError("Setting '%s' must be of type '%s'" % (name, klass))

        settings_item.validate(name, value)
        self._settings[name] = value

        LOG.debug(f"_set {name}={value} stack_size={len(self._stack)}")

    @forward
    def reset(self, name: str = None):
        """Reset setting(s) to default values.

        Parameters
        ----------
            name: str, optional
                The name of the setting to reset to default. If the setting
                does not have a default, it is removed. If `None` is passed, all settings are
                reset to their default values. Defaults to None.
        """
        if name is None:
            self._settings = dict(**DEFAULTS)
        else:
            if name not in DEFAULTS:
                raise KeyError("No setting name '%s'" % (name,))

            self._settings.pop(name, None)
            if name in DEFAULTS:
                self._settings[name] = DEFAULTS[name]
        self._changed()

    @forward
    def __repr__(self):
        r = []
        for k, v in sorted(self._settings.items()):
            setting = SETTINGS_AND_HELP.get(k, None)
            default = setting.default if setting else ""
            r.append(f"{k}: ({v}, {default})")
        return "\n".join(r)

    @forward
    def _repr_html_(self):
        html = [css("table")]
        html.append("<table class='ek'>")
        html.append("<tr><th>Name</th><th>Value</th><th>Default</th></tr>")
        for k, v in sorted(self._settings.items()):
            setting = SETTINGS_AND_HELP.get(k, None)
            default = setting.default if setting else ""
            html.append("<tr><td>%s</td><td>%r</td><td>%r</td></tr>" % (k, v, default))
        html.append("</table>")
        return "".join(html)

    @forward
    def dump(self):
        for k, v in sorted(self._settings.items()):
            yield ((k, v, SETTINGS_AND_HELP.get(k)))

    def _changed(self):
        if self._auto_save_settings:
            self._save()
        self._notify()

    def _notify(self):
        if self._notify_enabled:
            for cb in self._callbacks:
                cb()

    def on_change(self, callback: Callable[[], None]):
        self._callbacks.append(callback)

    def _save(self):
        if self._settings_yaml is None:
            return

        try:
            save_settings(self._settings_yaml, self._settings)
        except Exception:
            LOG.error(
                "Cannot save earthkit-data settings (%s)",
                self._settings_yaml,
                exc_info=True,
            )

    def save_as(self, path):
        try:
            save_settings(path, self._settings)
        except Exception:
            LOG.error(
                f"Cannot save earthkit-data settings ({path})",
                exc_info=True,
            )

    def _as_bytes(self, name, value, none_ok):
        return as_bytes(value, name=name, none_ok=none_ok)

    def _as_percent(self, name, value, none_ok):
        return as_percent(value, name=name, none_ok=none_ok)

    def _as_seconds(self, name, value, none_ok):
        return as_seconds(value, name=name, none_ok=none_ok)

    def _as_str(self, name, value, none_ok):
        if value is None and none_ok:
            return None
        return str(value)

    # def _as_number(self, name, value, units, none_ok):
    #     return as_number(name, value, units, none_ok)

    @forward
    def temporary(self, *args, **kwargs):
        tmp = Settings(None, self._settings)
        # until the tmp object is at the top of the stack we do not want
        # notify the observers
        if len(args) > 0 or len(kwargs) > 0:
            # tmp does not have any callbacks so it will not broadcast the changes
            tmp.set(*args, **kwargs)
        tmp._callbacks = self._callbacks
        return new_settings(tmp)

    @property
    def auto_save_settings(self):
        return Settings._auto_save_settings

    @auto_save_settings.setter
    def auto_save_settings(self, v):
        Settings._auto_save_settings = v


save = False
settings_yaml = os.path.expanduser(os.path.join(DOT_EARTHKIT_DATA, "settings.yaml"))

try:
    if not os.path.exists(DOT_EARTHKIT_DATA):
        os.mkdir(DOT_EARTHKIT_DATA, 0o700)
    if not os.path.exists(settings_yaml):
        save_settings(settings_yaml, DEFAULTS)
except Exception:
    LOG.error(
        "Cannot create earthkit-data settings directory, using defaults (%s)",
        settings_yaml,
        exc_info=True,
    )

settings = dict(**DEFAULTS)
try:
    with open(settings_yaml) as f:
        s = yaml.load(f, Loader=yaml.SafeLoader)
        if not isinstance(s, dict):
            s = {}

        settings.update(s)

    # if s != settings:
    #     save = True

    if settings.get("version") != VERSION:
        save = True

except Exception:
    LOG.error(
        "Cannot load earthkit-data settings (%s), reverting to defaults",
        settings_yaml,
        exc_info=True,
    )

SETTINGS = Settings(settings_yaml, settings)
if save:
    SETTINGS._save()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

_skip_methods = {
    "data.readers.bufr.bufr.BUFRList": [
        "cache_file",
        "dataset",
        "from_dict",
        "from_mask",
        "from_multi",
        "from_slice",
        "full",
        "graph",
        "ignore",
        "merge",
        "mutate",
        "new_mask_index",
        "parent",
        "settings",
        "to_numpy",
    ],
    "data.readers.bufr.bufr.BUFRMessage": ["merge", "mutate", "to_numpy"],
    "data.readers.grib.codes.GribField": ["merge", "mutate"],
    "data.core.fieldlist.FieldList": [
        "cache_file",
        "dataset",
        "from_dict",
        "from_mask",
        "from_multi",
        "from_slice",
        "full",
        "graph",
        "ignore",
        "merge",
        "mutate",
        "new_mask_index",
        "parent",
        "scaled",
        "settings",
        "statistics",
        "xarray_open_dataset_kwargs",
    ],
    "data.readers.grib.index.GribFieldList": [
        "cache_file",
        "dataset",
        "from_dict",
        "from_mask",
        "from_multi",
        "from_slice",
        "full",
        "graph",
        "ignore",
        "merge",
        "mutate",
        "new_mask_index",
        "parent",
        "scaled",
        "settings",
        "statistics",
        "xarray_open_dataset_kwargs",
    ],
}


# define skip rules for autoapi
def _skip_api_items(app, what, name, obj, skip, options):
    # if what == "package":
    #     print(f"{name}[{what}]")

    if what == "module" and name not in [
        "data.core",
        "data.core.metadata",
        "data.core.fieldlist",
        "data.readers",
        "data.readers.bufr.bufr",
        "data.readers.grib.codes",
        "data.readers.grib.index",
        "data.utils",
        "data.utils.bbox",
    ]:
        skip = True
    elif what == "package" and name not in [
        "data",
        "data.core",
        "data.readers",
        "data.readers.bufr",
        "data.readers.bufr.bufr",
        "data.readers.grib",
        "data.readers.grib.index",
        "data.utils",
        "data.utils.bbox",
    ]:
        skip = True
    elif what == "class" and name not in [
        "data.core.fieldlist.Field",
        "data.core.fieldlist.FieldList",
        "data.core.metadata.Metadata",
        "data.core.metadata.RawMetadata",
        "data.readers.bufr.bufr.BUFRList",
        "data.readers.bufr.bufr.BUFRMessage",
        "data.readers.grib.codes.GribField",
        "data.readers.grib.index.GribFieldList",
        "data.readers.grib.metadata.GribMetadata",
        "data.utils.bbox.BoundingBox",
    ]:
        skip = True
    elif what == "method":
        # if "abstractmethod" in getattr(obj, "properties", []):
        #     skip = True
        # else:
        if True:
            for k, v in _skip_methods.items():
                if k in name and name.split(".")[-1] in v:
                    skip = True
                    break

    elif what in ("function", "attribute", "data"):
        skip = True

    # if not skip:
    #     print(f"{what} {name}")
    return skip
